from typing import Dict, Union, Literal, Any, Annotated
from pydantic import (
    Field,
    field_validator,
    model_validator,
    computed_field,
    TypeAdapter,
)

import eole
from eole.constants import PositionEncodingType, ActivationFunction, ModelType
from eole.config.config import Config


class EmbeddingsConfig(Config):
    src_word_vec_size: int = Field(default=512, description="Word embedding size for src.")
    tgt_word_vec_size: int = Field(default=512, description="Word embedding size for tgt.")
    word_vec_size: int = Field(default=-1, description="Word embedding size for src and tgt.")
    # Freeze word vectors
    freeze_word_vecs_enc: bool = Field(default=False, description="Freeze word embeddings on the encoder side.")
    freeze_word_vecs_dec: bool = Field(default=False, description="Freeze word embeddings on the encoder side.")
    position_encoding: bool = Field(
        default=False,
        description="Absolute position encoding, see position_encoding_type. " "Necessary for non-RNN style models.",
    )
    position_encoding_type: PositionEncodingType | None = Field(
        default=PositionEncodingType.SinusoidalInterleaved,
        description="Type of positional encoding.",
    )
    n_positions: int | None = Field(
        default=None,
        description="Two cases"
        "Case 1: Absolute number of positions to learn "
        "position embeddings on (position_encoding_type: Learned)"
        "Case 2: Max Relative Positions"
        "In the case of position_encoding_type: Relative",
    )
    position_shift: int | None = Field(
        default=0,
        description="Positions IDS shift before making position embed " "dirty patch to cover for xlm-roberta-xl",
    )
    normalize: bool | None = Field(
        default=False,
        description="Enable embeddings scaling. "
        "Not always necessary, but useful for some model compatibility, e.g. gemma. "
        "https://datascience.stackexchange.com/a/87909",
    )

    @model_validator(mode="after")
    def validate_embeddings(self):
        if self.position_encoding_type in [
            PositionEncodingType.Learned,
            PositionEncodingType.Relative,
        ]:
            assert self.n_positions is not None, (
                "n_positions must be set if position_encoding_type "
                f"is {PositionEncodingType.Learned} or {PositionEncodingType.Relative}"
            )
        elif self.position_encoding_type in [
            PositionEncodingType.SinusoidalInterleaved,
            PositionEncodingType.SinusoidalConcat,
        ]:
            assert not (self.normalize), "embeddings normalization is already handled in PositionalEncoding"
        return self


class EncoderConfig(Config):
    """Abstract class for all encoders"""

    encoder_type: str | None = Field(default="rnn", description="Type of encoder layer(s) to use.")
    layers: int = Field(default=2, description="Number of layers in the encoder.")
    hidden_size: int = Field(default=512, description="Size of encoder hidden states.")

    # This field should be set at EmbeddingsConfig level but will be copied here for cases
    # where input size to the rnn is different to the hidden size
    src_word_vec_size: int = Field(default=512, description="Word embedding size for src.")


class DecoderConfig(Config):
    """Abstract class for all decoders"""

    decoder_type: str | None = Field(default="rnn", description="Type of decoder layer(s) to use.")
    layers: int = Field(default=2, description="Number of layers in the decoder.")
    hidden_size: int = Field(default=512, description="Size of decoder hidden states.")

    # This field should be set at EmbeddingsConfig level but will be copied here for cases
    # where input size to the rnn is different to the hidden size
    tgt_word_vec_size: int = Field(default=512, description="Word embedding size for tgt.")
    coverage_attn: bool = Field(default=False, description="Train a coverage attention layer.")
    lambda_coverage: float = Field(default=0.0, description="Lambda value for coverage loss of See et al (2017)")
    global_attention: Literal["dot", "general", "mlp", None] = Field(
        default="general",
        description="The attention type to use. (Luong=general, Bahdanau=MLP)",
    )
    global_attention_function: Literal["softmax", "sparsemax"] = Field(
        default="softmax", description="Global attention function to use."
    )

    @model_validator(mode="after")
    def _validate_decoder_config(self):
        return self


class RnnConfig(Config):
    """Just to facilitate testing discriminator stuff."""

    bridge: bool = Field(
        default=False,
        description="Have an additional layer between the last encoder state "
        "and the first decoder state (RNN specific).",
    )
    rnn_type: Literal["LSTM", "GRU"] = Field(
        default="LSTM", description="The gate type to use in the RNNs."
    )  # this should be homogenized with available encoder classes (use some registry? or at least a clean Enum somewhere) # noqa: E501


class RnnEncoderConfig(RnnConfig, EncoderConfig):
    encoder_type: Literal["rnn", "brnn"] = Field(default="rnn")


class RnnDecoderConfig(RnnConfig, DecoderConfig):
    decoder_type: Literal["rnn"] = Field(default="rnn")
    context_gate: Literal["source", "target", "both", None] = Field(
        default=None, description="Type of context gate to use."
    )
    bidirectional_encoder: bool | None = False


class CnnConfig(Config):
    cnn_kernel_width: int = Field(
        default=3,
        description="Size of windows in the cnn, the kernel_size is " "(cnn_kernel_width, 1) in convolution layers.",
    )


class CnnEncoderConfig(CnnConfig, EncoderConfig):
    encoder_type: Literal["cnn"] = Field(default="cnn")


class CnnDecoderConfig(CnnConfig, DecoderConfig):
    decoder_type: Literal["cnn"] = Field(default="cnn")


class MeanEncoderConfig(EncoderConfig):
    encoder_type: Literal["mean"] = Field(default="mean")


class RotaryPositionConfig(Config):
    """
    Configuration for rotary position embeddings used in transformer models.
    """

    rotary_interleave: bool = Field(
        default=True,
        description="Interleave the head dimensions when rotary embeddings are applied. "
        "Otherwise the head dimensions are sliced in half. "
        "(True= Llama from Meta (original), "
        "False= used by all HuggingFace models)",
    )
    rotary_theta: int = Field(
        default=10000,
        description="Rotary theta base length, 1e4 for Llama2.Mistral, 1e6 for Mixtral",
    )
    rotary_dim: int = Field(
        default=0,
        description="Rotary dim when model requires it to be different to head dim.",
    )
    scaling_type: str | None = Field(
        default=None,
        description="Specifies the type of RoPE scaling to be applied, if any.",
    )
    alpha: float | None = Field(None, description="alpha factor by which to scale rope theta.")
    scaling_factor: float | None = Field(default=8.0, description="Factor by which to scale RoPE embeddings.")
    low_freq_factor: float | None = Field(
        default=1.0,
        description="Scaling factor applied to the lower frequency components of RoPE.",
    )
    high_freq_factor: float | None = Field(
        default=4.0,
        description="Scaling factor applied to the higher frequency components of RoPE.",
    )
    original_max_position_embeddings: int | None = Field(
        default=8192,
        description="Original maximum position embeddings for RoPE scaling.",
    )
    rotary_theta_local: int = Field(
        default=10000,
        description="Rotary theta base length for local rotary layers",
    )
    interleave_local: int = Field(
        default=0,
        description="Local rotary layers each 1/N layers",
    )
    tmax_index: int = Field(
        default=0,
        description="tmax indexing, 0 for all cases except gemma 3 = 1",
    )


class TransformerConfig(Config):
    """
    This base TransformerConfig class regroups parameters than can
    both be set at model level or either encoder/decoder level.
    BaseModelConfig._override_values validator overrides
    encoder/decoder values with model values if relevant.
    """

    sliding_window: int = Field(default=0, description="Sliding window for transformer self-attention.")
    heads: int = Field(default=8, description="Number of heads for transformer self-attention.")
    transformer_ff: int = Field(default=2048, description="Size of hidden transformer feed-forward.")
    moe_transformer_ff: int | None = Field(default=None, description="Size of hidden moe transformer feed-forward.")
    relative_positions_buckets: int = Field(
        default=0,
        description="Enable relative position bias "
        "(https://github.com/google-research/text-to-text-transfer-transformer).",
    )
    mlp_activation_fn: ActivationFunction = Field(
        default=ActivationFunction.relu,
        description="The activation function to use in MLP layer.",
    )
    layer_norm: Literal["standard", "standardFP32", "rms", "gemma-rms"] = Field(
        default="standard",
        description="Type of layer normalization in transformer architecture.",
    )
    norm_eps: float = Field(default=1e-5, description="Layer norm epsilon.")
    shared_layer_norm: bool = Field(
        default=False,
        description="Use a shared layer_norm in parallel residual attention. "
        "Note: must be True for Falcon 7B, False for Falcon 40B, "
        "same for GPT-J and GPT-NeoX models.",
    )
    ffn_layernorm: bool = Field(
        default=False,
        description="Add pre/post_feedforward_layernorm around MLP forward. " "Note: introduced for gemma2 support.",
    )
    add_qkvbias: bool = Field(
        default=False,
        description="Add bias to nn.Linear of Query/Key/Value in MHA. "
        "Note: this will add bias to output projection layer too by default. "
        "Can be disabled with `add_final_linear_bias`.",
    )
    query_norm: bool = Field(
        default=False,
    )
    key_norm: bool = Field(
        default=False,
    )
    qk_norm_post_rope: bool = Field(
        default=False,
    )
    add_final_linear_bias: bool = Field(default=False, description="Add bias to nn.Linear of final_linear in MHA.")
    heads_kv: int | None = Field(
        default=None,
        description="Number of heads for KV. heads_kv=heads if None, else number of heads for KV" "(e.g. Falcon 40B)",
    )
    head_dim: int | None = Field(
        default=None,
        description="Head dimension when this needs to be different vs hidden_size // heads",
    )
    attn_scaling: float | None = Field(
        default=None,
        description="Attention scaling factor, when None uses 1/sqrt(head_dim) by default",
    )
    add_ffnbias: bool = Field(default=False, description="Add bias to nn.Linear of MLP FFN.")
    parallel_residual: bool = Field(
        default=False,
        description="Use parallel residual in decoder layer. " "Note: this is used by GPT-J / Falcon Architecture.",
    )
    num_experts: int = Field(default=0, description="Number of experts for MoE models.")
    num_shared_experts: int = Field(default=0, description="Number of shared experts for MoE models (DeepSeekv2).")
    first_k_dense_replace: int = Field(default=0, description="Number of layers using Dense instead of MoE")
    num_experts_per_tok: int = Field(default=2, description="Number of experts per token.")
    moe_softmax_after: bool = Field(default=False, description="Usually softmax is before topk, Mixtral does it after.")
    # These fields are set at EmbeddingsConfig level but will be copied here to be accessible in MHA
    position_encoding_type: PositionEncodingType | None = Field(
        default=PositionEncodingType.SinusoidalInterleaved,
        description="Type of positional encoding.",
    )
    interpolate_mode: str | None = Field(
        default=None,
        description=(
            "Interpolation mode for position embeddings. "
            "If None: position_embeddings is a lookup table based on n_positions. "
            "If string (e.g., 'bilinear'): position_embedding is a learned grid using interpolation."
            " (see Vision.py encoder)"
        ),
    )
    n_positions: int | None = Field(
        default=None,
        description="Two cases"
        "Case 1: Absolute number of positions to learn "
        "position embeddings on (position_encoding_type: Learned)"
        "Case 2: Max Relative Positions"
        "In the case of position_encoding_type: Relative",
    )
    rope_config: RotaryPositionConfig | None = Field(default=None, description="Rotary position config, if relevant.")

    @model_validator(mode="after")
    def _validate_transformer_config(self):

        if self.position_encoding_type == PositionEncodingType.Rotary:
            if self.rope_config is None:
                self.rope_config = RotaryPositionConfig()

        if self.add_qkvbias and "add_final_linear_bias" not in self.model_fields_set:
            self.update(add_final_linear_bias=True)
        return self

    @computed_field
    @property
    def dim_per_head(self) -> int:
        if self.head_dim is not None:
            return self.head_dim
        else:
            return self.hidden_size // self.heads


# could eole.encoders.TransformerEncoder class inherit from this? (it seems not unfortunately)
class TransformerEncoderConfig(TransformerConfig, EncoderConfig):
    encoder_type: Literal["transformer"] = Field(
        default="transformer"
    )  # not sure it's fully proper use, but inspired from docs -- https://docs.pydantic.dev/latest/concepts/fields/#discriminator # noqa: E501


class TransformerDecoderConfig(TransformerConfig, DecoderConfig):
    decoder_type: Literal["transformer"] = Field(default="transformer")

    aan_useffn: bool = Field(default=False, description="Turn on the FFN layer in the AAN decoder.")
    # some guided alignment specific opts were omitted here (put back for backcompat)
    alignment_layer: int = Field(default=-2, description="Layer number which has to be supervised.")
    alignment_heads: int = Field(
        default=0,
        description="Number of cross attention heads per layer to supervise with.",
    )
    full_context_alignment: bool = Field(
        default=False,
        description="Whether alignment is conditioned on full target context.",
    )
    lambda_align: float = Field(
        default=0.0,
        description="Lambda value for alignement loss of Garg et al, 2019 " "(https://arxiv.org/abs/1909.02074)",
    )
    LM_type: Literal["causal", "prefix"] = Field(
        default="causal",
        description="TransformerDecoder LM type (causal = classic, or prefix LM https://arxiv.org/pdf/2308.06912)",
    )

    @model_validator(mode="after")
    def _validate_transformer_decoder_config(self):
        # move to specific decoder config ?
        # this is not working right now if enc_layers is overridden by model level config
        # if self.lambda_align > 0.0:
        #     assert (
        #         self.decoder_type == "transformer"
        #     ), "Only transformer is supported to joint learn alignment."
        #     assert (
        #         self.alignment_layer < self.dec_layers
        #         and self.alignment_layer >= -self.dec_layers
        #     ), f"N° alignment_layer ({self.alignment_layer}) should be smaller "\
        #         "than number of layers ({self.dec_layers})."
        #     logger.info(
        #         "Joint learn alignment at layer [{}] "
        #         "with {} heads in full_context '{}'.".format(
        #             self.alignment_layer,
        #             self.alignment_heads,
        #             self.full_context_alignment,
        #         )
        #     )

        assert (
            self.hidden_size % self.heads == 0
        ), "Transformer Model dimension {} must be divisible by the number of heads {}".format(
            self.hidden_size, self.heads
        )
        return self


class VisionEncoderConfig(TransformerConfig, EncoderConfig):
    """
    Based on mistral-community/pixtral-12b, might evolve later.
    """

    encoder_type: Literal["vision"] = Field(default="vision")
    # default to Pixtral 12B settings, might change later
    num_channels: int | None = 3
    image_size: int | None = 1024
    patch_size: int | None = 16
    image_token_id: int | None = 10  # pixtral uses 10, gemma3 uses 262144
    mm_tokens_per_image: int | None = 256  # added for gemma3
    layernorm_pre: bool = True  # True for pixtral/mistral/deepseek False for gemma3/hunyuan
    layernorm_post: bool = False  # False for pixtral/mistral/deepseek/hunyuan True for gemma3
    patch_conv_bias: bool = False  # False for pixtral/mistral True for gemma3

    encoder_sam: bool = False  # True for DeepSeekOCR (Segment Anything Model as 1st step)
    use_class_embedding: bool = False


# use Field with default= + description would be more readable
# was inheriting from VocabConfig, but removed for now to facilitate inference tests
# could we have different BaseModelConfig classes (inheriting from a base one)
# for different architectures / model types?
class BaseModelConfig(Config):
    # some experiments around nested config
    embeddings: EmbeddingsConfig = Field(
        default_factory=EmbeddingsConfig,
        description="Contains most of the args useful to build the Embeddings module.",
    )
    encoder: (
        Union[
            TransformerEncoderConfig,
            RnnEncoderConfig,
            CnnEncoderConfig,
            MeanEncoderConfig,
            VisionEncoderConfig,
        ]
        | None
    ) = Field(
        default=None,
        discriminator="encoder_type",
        description="Major parameters of an encoder.",
    )  # we shall use discriminators here
    decoder: (
        Union[
            TransformerDecoderConfig,
            RnnDecoderConfig,
            CnnDecoderConfig,
        ]
        | None
    ) = Field(
        default=None,
        discriminator="decoder_type",
        description="Major parameters of a decoder.",
    )  # we shall use discriminators here

    # model level stuff that may override enc/dec level
    hidden_size: int = Field(
        default=-1,
        description="Size of hidden states. Overwrites [encoder/decoder].hidden_size if set.",
    )
    word_vec_size: int = Field(default=-1, description="Word embedding size for src and tgt.")
    layers: int = Field(
        default=-1,
        description="Number of layers in both encoder and decoder " "(will overwrite enc_layers/dec_layers).",
    )
    transformer_ff: int = Field(default=-1, description="Size of hidden transformer feed-forward.")
    moe_transformer_ff: int | None = Field(default=None, description="Size of hidden moe transformer feed-forward.")

    share_decoder_embeddings: bool = Field(
        default=False,
        description="Use a share weight matrix for the input and output " "word embeddings in the decoder.",
    )
    share_embeddings: bool = Field(
        default=False,
        description="Share the word embeddings between encoder and decoder. "
        "Need to use shared vocabulary for this option.",
    )
    input_feed: int = Field(
        default=1,
        description="Feed the context vector at each time step as additional input "
        "(via concatenation with the word embeddings) to the decoder.",
    )  # is this useful / working anymore?
    generator_function: Literal["softmax", "sparsemax"] = Field(
        default="softmax",
        description="Which function to use for generating probabilities " "over the target vocabulary.",
    )
    generator_bias: bool = Field(
        default=True,
        description="Control whether or not the generator Linear module has bias weights.",
    )
    adapter_bias: bool = Field(
        default=False,
        description="Control whether or not the adapter module has bias weights.",
    )
    projector_activation_fn: ActivationFunction | None = Field(
        default=ActivationFunction.relu,
        description="The activation function to use in adapter projector layer.",
    )
    spatial_merge_size: int | None = Field(
        default=1,
        description="Control the presence and size of patch merger (Mistral3)",
    )

    add_estimator: bool = Field(default=False, description="Add estimator layer")
    estimator_type: Literal["average", "last_token", "first_token"] = Field(
        default="average",
        description="Which hidden_states to use to feed the estimator",
    )

    left_pad: bool = Field(default=False, description="Enable left-padding, useful for some LLMs.")
    huggingface_model: str | None = Field(default=None, description="Original huggingface model.")
    eole_version: str | None = Field(
        default=eole.__version__,
        description="Eole version used to convert/train/save the model.",
    )

    # @computed_field()
    # @property
    # def brnn(self) -> bool:
    #     if self.encoder is not None:
    #         return self.encoder.encoder_type == "brnn"

    @field_validator("spatial_merge_size")
    @classmethod
    def validate_merge_size(cls, v: int | None) -> int:
        return 1 if v is None else v

    @property
    def model_type(self) -> ModelType:
        if self.decoder is None:
            return ModelType.ENCODER
        elif self.encoder is None:
            return ModelType.DECODER
        else:
            return ModelType.ENCODER_DECODER

    @field_validator("encoder", "decoder", "embeddings", mode="before")
    @classmethod
    def str_to_dict(cls, v) -> Dict:
        if isinstance(v, str):
            # not sure about this patch to allow forcing encoder/decoder to None
            if v == "None":
                return None
            w = eval(v)
            return w
        return v

    # @field_validator("decoder", mode="before")
    # @classmethod
    # def str_to_dict(cls, v) -> Dict:
    #     if isinstance(v, str):
    #         return eval(v)
    #     return v

    architecture: str | None = None

    @model_validator(mode="before")
    @classmethod
    def default_architecture(cls, data: Any) -> Any:
        if not (isinstance(data, dict)):
            return data
        if "architecture" not in data.keys():
            data["architecture"] = "custom"
        return data

    def update_model_opts(self):
        update_dict = {}
        """
        if self.embeddings.position_encoding_type == PositionEncodingType.Rotary:
            if not self.rope_config:
                update_dict["rope_config"] = RotaryPositionConfig()
                rope_config = update_dict["rope_config"]
            else:
                rope_config = self.rope_config
        else:
            rope_config = None
        """
        if self.embeddings is not None and self.embeddings.word_vec_size > 0:
            update_dict["embeddings"] = {
                "src_word_vec_size": self.embeddings.word_vec_size,
                "tgt_word_vec_size": self.embeddings.word_vec_size,
            }
        if self.embeddings is not None and "embeddings" in update_dict.keys():
            self.embeddings.update(**update_dict.pop("embeddings"))

        if getattr(self.encoder, "encoder_type", None) == "brnn" and self.decoder.decoder_type == "rnn":
            update_dict["decoder"] = {"bidirectional_encoder": True}

        if self.encoder is not None and hasattr(self.encoder, "src_word_vec_size"):
            update_dict["encoder"] = {"src_word_vec_size": self.embeddings.src_word_vec_size}
            if getattr(self.encoder, "encoder_type", None) == "transformer":
                update_dict["encoder"].update(
                    {
                        "position_encoding_type": self.embeddings.position_encoding_type,
                        "n_positions": self.embeddings.n_positions,
                        # "rope_config": rope_config,
                    }
                )
                update_dict["position_encoding_type"] = self.embeddings.position_encoding_type
        if self.encoder is not None and "encoder" in update_dict.keys():
            self.encoder.update(**update_dict.pop("encoder"))

        if self.decoder is not None:
            update_dict["decoder"] = {"tgt_word_vec_size": self.embeddings.tgt_word_vec_size}
            if getattr(self.decoder, "decoder_type", None) == "transformer":
                update_dict["decoder"].update(
                    {
                        "position_encoding_type": self.embeddings.position_encoding_type,
                        "n_positions": self.embeddings.n_positions,
                        # "rope_config": rope_config,
                    }
                )
                update_dict["position_encoding_type"] = self.embeddings.position_encoding_type
        if self.decoder is not None and "decoder" in update_dict.keys():
            self.decoder.update(**update_dict.pop("decoder"))

        self.update(**update_dict)

        # causing some weird recursion issue in unit test, to investigate
        # if self.encoder is not None:
        #     self.brnn = self.encoder.encoder_type == "brnn"

    @model_validator(mode="after")
    def _override_values(self):
        # aggregate all fields in a single dict,
        # and update once to prevent validation mismatch
        embeddings_fields = {}
        encoder_fields = {}
        decoder_fields = {}
        for field in self.model_fields_set:
            if hasattr(self.embeddings, field) and field not in self.embeddings.model_fields_set:
                embeddings_fields[field] = getattr(self, field)
            if hasattr(self.encoder, field) and field not in self.encoder.model_fields_set:
                encoder_fields[field] = getattr(self, field)
            if hasattr(self.decoder, field) and field not in self.decoder.model_fields_set:
                decoder_fields[field] = getattr(self, field)
        if self.embeddings is not None:
            self.embeddings.update(**embeddings_fields)
        if self.encoder is not None:
            self.encoder.update(**encoder_fields)
        if self.decoder is not None:
            self.decoder.update(**decoder_fields)
        return self

    @model_validator(mode="after")
    def _validate_model_config(self):
        self.update_model_opts()

        # encoder and decoder should be same sizes
        if self.encoder is not None and self.decoder is not None:
            if isinstance(self.encoder, VisionEncoderConfig):
                pass
            else:
                same_size = self.encoder.hidden_size == self.decoder.hidden_size
                assert same_size, "The encoder and decoder must have the same hidden size for seq2seq models."

        if self.share_embeddings:
            if self.encoder is None or self.decoder is None:
                raise AssertionError("--share_embeddings is for EncoderDecoder models only.")

        return self


class CustomModelConfig(TransformerConfig, BaseModelConfig):
    """
    Wrap anything that does not fit a set common architecture.
    """

    architecture: Literal["custom"] = Field(default="custom")


class RnnModelConfig(RnnConfig, BaseModelConfig):
    architecture: Literal["rnn"] = Field(default="rnn")

    @model_validator(mode="before")
    @classmethod
    def encoder_decoder_type(cls, data: Any) -> Any:
        # patch to allow transparent setting of encoder/decoder_type
        if not (isinstance(data, dict)):
            return data
        if "encoder" in data.keys():
            data["encoder"]["encoder_type"] = "rnn"
        else:
            data["encoder"] = {"encoder_type": "rnn"}
        if "decoder" in data.keys():
            data["decoder"]["decoder_type"] = "rnn"
        else:
            data["decoder"] = {"decoder_type": "rnn"}
        return data

    @model_validator(mode="before")
    @classmethod
    def default_architecture(cls, data: Any) -> Any:
        if not (isinstance(data, dict)):
            return data
        if "architecture" not in data.keys():
            data["architecture"] = "rnn"
        return data


class CnnModelConfig(CnnConfig, BaseModelConfig):
    architecture: Literal["cnn"] = Field(default="cnn")

    @model_validator(mode="before")
    @classmethod
    def encoder_decoder_type(cls, data: Any) -> Any:
        # we shall probably factorize this in BaseModelConfig
        # patch to allow transparent setting of encoder/decoder_type
        if not (isinstance(data, dict)):
            return data
        if "encoder" in data.keys():
            data["encoder"]["encoder_type"] = "cnn"
        else:
            data["encoder"] = {"encoder_type": "cnn"}
        if "decoder" in data.keys():
            data["decoder"]["decoder_type"] = "cnn"
        else:
            data["decoder"] = {"decoder_type": "cnn"}
        return data

    @model_validator(mode="before")
    @classmethod
    def default_architecture(cls, data: Any) -> Any:
        if not (isinstance(data, dict)):
            return data
        if "architecture" not in data.keys():
            data["architecture"] = "cnn"
        return data


class TransformerModelConfig(TransformerConfig, BaseModelConfig):
    """
    Facilitate setting some transformer specific params at model level.
    """

    architecture: Literal["transformer"] = Field(default="transformer")

    @model_validator(mode="before")
    @classmethod
    def encoder_decoder_type(cls, data: Any) -> Any:
        # patch to allow transparent setting of encoder/decoder_type
        if not (isinstance(data, dict)):
            return data
        if "encoder" in data.keys():
            data["encoder"]["encoder_type"] = "transformer"
        else:
            data["encoder"] = {"encoder_type": "transformer"}
        if "decoder" in data.keys():
            data["decoder"]["decoder_type"] = "transformer"
        else:
            data["decoder"] = {"decoder_type": "transformer"}
        return data

    @model_validator(mode="before")
    @classmethod
    def default_architecture(cls, data: Any) -> Any:
        if not (isinstance(data, dict)):
            return data
        if "architecture" not in data.keys():
            data["architecture"] = "transformer"
        return data

    @model_validator(mode="after")
    def _validate_transformer(self):
        return self


class TransformerLMModelConfig(TransformerConfig, BaseModelConfig):
    """
    Facilitate setting some transformer specific params at model level.
    """

    architecture: Literal["transformer_lm"] = Field(default="transformer_lm")

    # force encoder to None in TransformerLM case
    encoder: None = Field(default=None, description="Major parameters of an encoder.")

    @model_validator(mode="before")
    @classmethod
    def encoder_decoder_type(cls, data: Any) -> Any:
        # patch to allow transparent setting of encoder/decoder_type
        if not (isinstance(data, dict)):
            return data
        if "decoder" in data.keys():
            data["decoder"]["decoder_type"] = "transformer"
        else:
            data["decoder"] = {"decoder_type": "transformer"}
        return data

    @model_validator(mode="before")
    @classmethod
    def default_architecture(cls, data: Any) -> Any:
        if not (isinstance(data, dict)):
            return data
        if "architecture" not in data.keys():
            data["architecture"] = "transformer_lm"
        return data

    @model_validator(mode="after")
    def _validate_transformer(self):
        # duplicate with TransformerModelConfig, might merge at some point
        return self


class VisionTransformerLMModelConfig(TransformerConfig, BaseModelConfig):
    architecture: Literal["vision_transformer_lm"] = Field(default="vision_transformer_lm")

    adapter: str | None = Field(default="llava", description="Adapter type to use in the model.")

    @model_validator(mode="before")
    @classmethod
    def encoder_decoder_type(cls, data: Any) -> Any:
        # patch to allow transparent setting of encoder/decoder_type
        if not (isinstance(data, dict)):
            return data
        if "encoder" in data.keys():
            data["encoder"]["encoder_type"] = "vision"
        else:
            data["encoder"] = {"encoder_type": "vision"}
        if "decoder" in data.keys():
            data["decoder"]["decoder_type"] = "transformer"
        else:
            data["decoder"] = {"decoder_type": "transformer"}
        return data

    @model_validator(mode="before")
    @classmethod
    def default_architecture(cls, data: Any) -> Any:
        if not (isinstance(data, dict)):
            return data
        if "architecture" not in data.keys():
            data["architecture"] = "vision_transformer_lm"
        return data

    @model_validator(mode="after")
    def _validate_vision_transformer(self):
        # assert not (self.add_estimator), "Estimator layer not supported in Vision Transformer"
        return self

    @property
    def image_size(self):
        return self.encoder.image_size

    @property
    def patch_size(self):
        return self.encoder.patch_size


class TransformerEncoderModelConfig(TransformerConfig, BaseModelConfig):
    """
    Facilitate setting some transformer specific params at model level.
    """

    architecture: Literal["transformer_encoder"] = Field(default="transformer_encoder")

    # force decoder to None in TransformerEncoder case
    decoder: None = Field(default=None, description="Major parameters of a decoder.")

    @model_validator(mode="before")
    @classmethod
    def encoder_decoder_type(cls, data: Any) -> Any:
        # patch to allow transparent setting of encoder/decoder_type
        if not (isinstance(data, dict)):
            return data
        if "encoder" in data.keys():
            data["encoder"]["encoder_type"] = "transformer"
        else:
            data["encoder"] = {"encoder_type": "transformer"}
        return data

    @model_validator(mode="before")
    @classmethod
    def default_architecture(cls, data: Any) -> Any:
        if not (isinstance(data, dict)):
            return data
        if "architecture" not in data.keys():
            data["architecture"] = "transformer_encoder"
        return data

    @model_validator(mode="after")
    def _validate_transformer(self):
        # duplicate with TransformerModelConfig, might merge at some point
        if getattr(self.embeddings, "position_encoding", False) and self.max_relative_positions != 0:
            raise ValueError(
                "Cannot use absolute and relative position encoding at the"
                "same time. Use either --position_encoding=true for legacy"
                "absolute position encoding or --max_realtive_positions with"
                " -1 for Rotary, or > 0 for Relative Position Representations"
                "as in https://arxiv.org/pdf/1803.02155.pdf"
            )
        return self


# Facilitate transparent instanciation from dumped fields
# https://stackoverflow.com/a/76538571
ModelConfig = Annotated[
    Union[
        TransformerModelConfig,
        TransformerLMModelConfig,
        VisionTransformerLMModelConfig,
        TransformerEncoderModelConfig,
        RnnModelConfig,
        CnnModelConfig,
        CustomModelConfig,
    ],
    Field(discriminator="architecture", default_factory=RnnModelConfig),  # noqa: F821
]

build_model_config = TypeAdapter(ModelConfig).validate_python



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import copy
import unittest
import glob
import os

from eole.train_single import prepare_transforms_vocabs
from eole.constants import CorpusName


SAVE_DATA_PREFIX = "eole/tests/data/test_data_prepare"


def get_default_opts():
    import yaml

    with open("eole/tests/data/data.yaml") as f:
        data_config = yaml.safe_load(f)["data"]
    # TODO: handle "config" override in pydantic model
    default_opts = {
        "data": data_config,
        "src_vocab": "eole/tests/data/vocab-train.src",
        "tgt_vocab": "eole/tests/data/vocab-train.tgt",
    }
    from eole.config.data import DataConfig

    opt = DataConfig(**default_opts)
    opt._validate_data_config()
    # Inject some dummy training options that may needed when build fields
    return opt


default_opts = get_default_opts()


class TestData(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super(TestData, self).__init__(*args, **kwargs)
        self.opt = default_opts

    def dataset_build(self, opt):
        try:
            prepare_transforms_vocabs(opt, {})
        except SystemExit as err:
            print(err)
        except IOError as err:
            if opt.skip_empty_level != "error":
                raise err
            else:
                print(f"Catched IOError: {err}")
        finally:
            # Remove the generated *pt files.
            for pt in glob.glob(SAVE_DATA_PREFIX + "*.pt"):
                os.remove(pt)
            if self.opt.save_data:
                # Remove the generated data samples
                sample_path = os.path.join(os.path.dirname(self.opt.save_data), CorpusName.SAMPLE)
                if os.path.exists(sample_path):
                    for f in glob.glob(sample_path + "/*"):
                        os.remove(f)
                    os.rmdir(sample_path)


def _add_test(param_setting, methodname):
    """
    Adds a Test to TestData according to settings

    Args:
        param_setting: list of tuples of (param, setting)
        methodname: name of the method that gets called
    """

    def test_method(self):
        if param_setting:
            opt = copy.deepcopy(self.opt)
            for param, setting in param_setting:
                setattr(opt, param, setting)
        else:
            opt = self.opt
        getattr(self, methodname)(opt)

    if param_setting:
        name = "test_" + methodname + "_" + "_".join(str(param_setting).split())
    else:
        name = "test_" + methodname + "_standard"
    setattr(TestData, name, test_method)
    test_method.__name__ = name


test_databuild = [
    [],
    [("src_vocab_size", 1), ("tgt_vocab_size", 1)],
    [("src_vocab_size", 10000), ("tgt_vocab_size", 10000)],
    # [("src_seq_len", 1)], # *_seq_len do not exist anymore it seems
    # [("src_seq_len", 5000)],
    # [("src_seq_length_trunc", 1)], # moved to transform opts, out of data config scope
    # [("src_seq_length_trunc", 5000)],
    # [("tgt_seq_len", 1)],
    # [("tgt_seq_len", 5000)],
    # [("tgt_seq_length_trunc", 1)],
    # [("tgt_seq_length_trunc", 5000)],
    [("share_vocab", True)],
    [("save_data", SAVE_DATA_PREFIX), ("n_sample", 30)],
    [("save_data", SAVE_DATA_PREFIX), ("n_sample", 30), ("skip_empty_level", "error")],
]

for p in test_databuild:
    _add_test(p, "dataset_build")



"""Define RNN-based encoders."""

import torch.nn as nn
import torch.nn.functional as F

from eole.encoders.encoder import EncoderBase


class RNNEncoder(EncoderBase):
    """A generic recurrent neural network encoder.

    Args:
        encoder_config (eole.config.ModelConfig)
        running_config (TrainingConfig / InferenceConfig)
    """

    def __init__(self, encoder_config, running_config=None):
        super(RNNEncoder, self).__init__()

        bidirectional = encoder_config.encoder_type == "brnn"

        num_directions = 2 if bidirectional else 1
        assert encoder_config.hidden_size % num_directions == 0
        hidden_size = encoder_config.hidden_size // num_directions

        self.rnn = getattr(nn, encoder_config.rnn_type)(
            batch_first=True,
            input_size=encoder_config.src_word_vec_size,
            hidden_size=hidden_size,
            num_layers=encoder_config.layers,
            dropout=getattr(running_config, "dropout", [0.0])[0],
            bidirectional=bidirectional,
        )

        # Initialize the bridge layer
        self.use_bridge = encoder_config.bridge
        if self.use_bridge:
            self._initialize_bridge(encoder_config.rnn_type, hidden_size, encoder_config.layers)

    @classmethod
    def from_config(cls, encoder_config, running_config=None):
        """Alternate constructor."""
        return cls(encoder_config, running_config=running_config)

    def forward(self, emb, **kwargs):
        """See :func:`EncoderBase.forward()`"""

        enc_out, enc_final_hs = self.rnn(emb)

        if self.use_bridge:
            enc_final_hs = self._bridge(enc_final_hs)

        return enc_out, enc_final_hs

    def _initialize_bridge(self, rnn_type, hidden_size, num_layers):
        # LSTM has hidden and cell state, other only one
        number_of_states = 2 if rnn_type == "LSTM" else 1
        # Total number of states
        self.total_hidden_dim = hidden_size * num_layers

        # Build a linear layer for each
        self.bridge = nn.ModuleList(
            [nn.Linear(self.total_hidden_dim, self.total_hidden_dim, bias=True) for _ in range(number_of_states)]
        )

    def _bridge(self, hidden):
        """Forward hidden state through bridge.
        final hidden state ``(num_layers x dir, batch, hidden_size)``
        """

        def bottle_hidden(linear, states):
            """
            Transform from 3D to 2D, apply linear and return initial size
            """
            states = states.permute(1, 0, 2).contiguous()
            size = states.size()
            result = linear(states.view(-1, self.total_hidden_dim))
            result = F.relu(result).view(size)
            return result.permute(1, 0, 2).contiguous()

        if isinstance(hidden, tuple):  # LSTM
            outs = tuple([bottle_hidden(layer, hidden[ix]) for ix, layer in enumerate(self.bridge)])
        else:
            outs = bottle_hidden(self.bridge[0], hidden)
        return outs

    def update_dropout(self, dropout, attention_dropout=None):
        self.rnn.dropout = dropout



import numpy as np
from PIL import Image
from typing import Optional, Union
from enum import Enum
from collections.abc import Collection
from eole.utils.logging import logger


DATASET_MEAN = {
    "llava": (0.48145466, 0.4578275, 0.40821073),
    "gemma3": (0.5, 0.5, 0.5),
    "deepseekocr": (0.5, 0.5, 0.5),
    "hunyuanocr": (0.48145466, 0.4578275, 0.40821073),
}
DATASET_STD = {
    "llava": (0.26862954, 0.26130258, 0.27577711),
    "gemma3": (0.5, 0.5, 0.5),
    "deepseekocr": (0.5, 0.5, 0.5),
    "hunyuanocr": (0.26862954, 0.26130258, 0.27577711),
}
RESAMPLE = {
    "llava": Image.BICUBIC,
    "gemma3": Image.BILINEAR,
    "deepseekocr": Image.BILINEAR,
    "hunyuanocr": Image.BILINEAR,
}
SQUARE = {"llava": False, "gemma3": True, "deepseekocr": True, "hunyuanocr": False}


def _convert_to_rgb(image: Image.Image) -> Image.Image:
    """
    Convert a PIL image to RGB.
    We ensure transparent background becomes white.
    """
    if image.mode == "RGB":
        return image
    if image.mode != "RGBA":
        image = image.convert("RGBA")
    white_bg: Image.Image = Image.new("RGBA", image.size, "WHITE")
    white_bg.paste(image, (0, 0), image)
    return white_bg.convert("RGB")


def image_to_num_tokens(img, image_size=0, image_patch_size=16, square=False):
    if square:  # Gemma3 (896x896) and Deepseekocr (1024x1024) require square fixed size
        return image_size // image_patch_size, image_size // image_patch_size
    w, h = img.size
    if image_size == 0 or image_size is None:
        image_size = max(w, h)
    ratio = max(h / image_size, w / image_size)
    if ratio > 1:
        w = round(w / ratio)
        h = round(h / ratio)
    width_tokens = (w - 1) // image_patch_size + 1
    height_tokens = (h - 1) // image_patch_size + 1
    # make numbers even to ensure batch even number of img tokens
    width_tokens = width_tokens + (width_tokens % 2)
    height_tokens = height_tokens + (height_tokens % 2)
    return width_tokens, height_tokens


class ChannelDimension(Enum):
    FIRST = "channels_first"
    LAST = "channels_last"


def infer_channel_dimension_format(
    image: np.ndarray, num_channels: Optional[Union[int, tuple[int, ...]]] = None
) -> ChannelDimension:
    """
    Infers the channel dimension format of `image`.

    Args:
        image (`np.ndarray`):
            The image to infer the channel dimension of.
        num_channels (`int` or `Tuple[int, ...]`, *optional*, defaults to `(1, 3)`):
            The number of channels of the image.

    Returns:
        The channel dimension of the image.
    """
    num_channels = num_channels if num_channels is not None else (1, 3)
    num_channels = (num_channels,) if isinstance(num_channels, int) else num_channels

    if image.ndim == 3:
        first_dim, last_dim = 0, 2
    elif image.ndim == 4:
        first_dim, last_dim = 1, 3
    else:
        raise ValueError(f"Unsupported number of image dimensions: {image.ndim}")

    if image.shape[first_dim] in num_channels and image.shape[last_dim] in num_channels:
        logger.warning(
            f"The channel dimension is ambiguous. Got image shape {image.shape}."
            f"Assuming channels are the first dimension."
        )
        return ChannelDimension.FIRST
    elif image.shape[first_dim] in num_channels:
        return ChannelDimension.FIRST
    elif image.shape[last_dim] in num_channels:
        return ChannelDimension.LAST
    raise ValueError("Unable to infer channel dimension format")


def rescale(
    image: np.ndarray,
    scale: float,
    data_format: Optional[ChannelDimension] = None,
    dtype: np.dtype = np.float32,
    input_data_format: Optional[Union[str, ChannelDimension]] = None,
) -> np.ndarray:
    """
    Rescales `image` by `scale`.

    Args:
        image (`np.ndarray`):
            The image to rescale.
        scale (`float`):
            The scale to use for rescaling the image.
        data_format (`ChannelDimension`, *optional*):
            The channel dimension format of the image. If not provided, it will be the same as the input image.
        dtype (`np.dtype`, *optional*, defaults to `np.float32`):
            The dtype of the output image. Defaults to `np.float32`. Used for backwards compatibility with feature
            extractors.
        input_data_format (`ChannelDimension`, *optional*):
            The channel dimension format of the input image. If not provided, it will be inferred from the input image.

    Returns:
        `np.ndarray`: The rescaled image.
    """
    if not isinstance(image, np.ndarray):
        raise TypeError(f"Input image must be of type np.ndarray, got {type(image)}")

    rescaled_image = image.astype(np.float64) * scale  # Numpy type promotion has changed, so always upcast first
    if data_format is not None:
        rescaled_image = to_channel_dimension_format(rescaled_image, data_format, input_data_format)

    rescaled_image = rescaled_image.astype(dtype)  # Finally downcast to the desired dtype at the end

    return rescaled_image


def get_channel_dimension_axis(
    image: np.ndarray, input_data_format: Optional[Union[ChannelDimension, str]] = None
) -> int:
    """
    Returns the channel dimension axis of the image.

    Args:
        image (`np.ndarray`):
            The image to get the channel dimension axis of.
        input_data_format (`ChannelDimension` or `str`, *optional*):
            The channel dimension format of the image. If `None`, will infer the channel dimension from the image.

    Returns:
        The channel dimension axis of the image.
    """
    if input_data_format is None:
        input_data_format = infer_channel_dimension_format(image)
    if input_data_format == ChannelDimension.FIRST:
        return image.ndim - 3
    elif input_data_format == ChannelDimension.LAST:
        return image.ndim - 1
    raise ValueError(f"Unsupported data format: {input_data_format}")


def normalize(
    image: np.ndarray,
    mean: Union[float, Collection[float]],
    std: Union[float, Collection[float]],
    data_format: Optional[ChannelDimension] = None,
    input_data_format: Optional[Union[str, ChannelDimension]] = None,
) -> np.ndarray:
    """
    Normalizes `image` using the mean and standard deviation specified by `mean` and `std`.

    image = (image - mean) / std

    Args:
        image (`np.ndarray`):
            The image to normalize.
        mean (`float` or `Collection[float]`):
            The mean to use for normalization.
        std (`float` or `Collection[float]`):
            The standard deviation to use for normalization.
        data_format (`ChannelDimension`, *optional*):
            The channel dimension format of the output image. If unset, will use the inferred format from the input.
        input_data_format (`ChannelDimension`, *optional*):
            The channel dimension format of the input image. If unset, will use the inferred format from the input.
    """
    if not isinstance(image, np.ndarray):
        raise ValueError("image must be a numpy array")

    if input_data_format is None:
        input_data_format = infer_channel_dimension_format(image)

    channel_axis = get_channel_dimension_axis(image, input_data_format=input_data_format)
    num_channels = image.shape[channel_axis]

    # We cast to float32 to avoid errors that can occur when subtracting uint8 values.
    # We preserve the original dtype if it is a float type to prevent upcasting float16.
    if not np.issubdtype(image.dtype, np.floating):
        image = image.astype(np.float32)

    if isinstance(mean, Collection):
        if len(mean) != num_channels:
            raise ValueError(f"mean must have {num_channels} elements if it is an iterable, got {len(mean)}")
    else:
        mean = [mean] * num_channels
    mean = np.array(mean, dtype=image.dtype)

    if isinstance(std, Collection):
        if len(std) != num_channels:
            raise ValueError(f"std must have {num_channels} elements if it is an iterable, got {len(std)}")
    else:
        std = [std] * num_channels
    std = np.array(std, dtype=image.dtype)

    if input_data_format == ChannelDimension.LAST:
        image = (image - mean) / std
    else:
        image = ((image.T - mean) / std).T

    image = to_channel_dimension_format(image, data_format, input_data_format) if data_format is not None else image
    return image


def to_channel_dimension_format(
    image: np.ndarray,
    channel_dim: Union[ChannelDimension, str],
    input_channel_dim: Optional[Union[ChannelDimension, str]] = None,
) -> np.ndarray:
    """
    Converts `image` to the channel dimension format specified by `channel_dim`.

    Args:
        image (`numpy.ndarray`):
            The image to have its channel dimension set.
        channel_dim (`ChannelDimension`):
            The channel dimension format to use.
        input_channel_dim (`ChannelDimension`, *optional*):
            The channel dimension format of the input image. If not provided, it will be inferred from the input image.

    Returns:
        `np.ndarray`: The image with the channel dimension set to `channel_dim`.
    """
    if not isinstance(image, np.ndarray):
        raise TypeError(f"Input image must be of type np.ndarray, got {type(image)}")

    if input_channel_dim is None:
        input_channel_dim = infer_channel_dimension_format(image)

    target_channel_dim = ChannelDimension(channel_dim)
    if input_channel_dim == target_channel_dim:
        return image

    if target_channel_dim == ChannelDimension.FIRST:
        image = image.transpose((2, 0, 1))
    elif target_channel_dim == ChannelDimension.LAST:
        image = image.transpose((1, 2, 0))
    else:
        raise ValueError("Unsupported channel dimension format: {}".format(channel_dim))
    return image


def process_image(image_path, adapter="llava", image_patch_size=16, image_size=1024):
    if isinstance(image_path, Image.Image):
        PILimage = image_path
    else:
        PILimage = Image.open(image_path)

    PILimage = _convert_to_rgb(PILimage)
    if adapter == "deepseekocr":
        image_patch_size = 16  # hardcoded (vs 14 for Clip module=encoder_config.patch_size)

    w, h = image_to_num_tokens(
        PILimage, image_size=image_size, image_patch_size=image_patch_size, square=SQUARE[adapter]
    )
    new_image_size = (w * image_patch_size, h * image_patch_size)
    PILimage = PILimage.resize(new_image_size, resample=RESAMPLE[adapter], reducing_gap=None)

    NPimage = np.array(PILimage, dtype=np.uint8)  # (H, W, C)
    NPimage = rescale(image=NPimage, scale=1 / 255, input_data_format=ChannelDimension.LAST)
    NPimage = normalize(
        image=NPimage,
        mean=DATASET_MEAN[adapter],
        std=DATASET_STD[adapter],
        input_data_format=ChannelDimension.LAST,
    )
    NPimage = to_channel_dimension_format(NPimage, ChannelDimension.FIRST, input_channel_dim=ChannelDimension.LAST)

    if adapter == "llava":
        image_tokens = (["[IMG]"] * w + ["[IMG_BREAK]"]) * h
        image_tokens[-1] = "[IMG_END]"
    elif adapter == "gemma3":
        image_tokens = "<start_of_image>" + "<image_soft_token>" * 256 + "<end_of_image>"
    elif adapter == "deepseekocr":
        # sam patches are 16 (hardcoded in original implementation)
        # for a 1024x1024 image it means patches of 64x64
        # for the sam encoder it is 64x64 = 4096 patches
        # then goes to Clip 4096 / 16 = 256 tokens (16 patches x 16 patches)
        # but there is one token at the end of each line, and one more at the end of the pic
        # (16 + 1) x 16 + 1 = 273
        image_tokens = ["<image>"] * 273
    elif adapter == "hunyuanocr":
        image_tokens = (
            "<｜hy_place▁holder▁no▁100｜>"
            + "<｜hy_place▁holder▁no▁102｜>" * (w * (h + 1) + 2)
            + "<｜hy_place▁holder▁no▁101｜>"
        )
    else:
        raise ValueError("Unsupported Adapter type: {}".format(adapter))
    return {"image": NPimage, "tokens": image_tokens}



from typing import List, Literal
from pydantic import Field, field_validator, model_validator, computed_field
from functools import cached_property
import torch

from eole.config.config import Config, get_config_dict
from eole.config.common import RunningConfig, LoRaConfig, QuantizeConfig
from eole.utils.logging import logger


class OptimizerConfig(Config):
    """
    Everything related to optimizers.
    Might be split into multiple subclasses later.
    Note: not fully sufficient (yet) to replace full opt namespace in build_torch_optimizer.
    Some other parameters (hidden_size, compute_dtype, etc.) are accessed.
    """

    optim: Literal[
        "sgd",
        "adagrad",
        "adadelta",
        "adam",
        "adamw",
        "sparseadam",
        "adafactor",
        "adamw8bit",
        "pagedadamw8bit",
        "pagedadamw32bit",
    ] = Field(
        default="sgd", description="Optimization method."
    )  # find a better way to list accepted values...
    adagrad_accumulator_init: float = Field(
        default=0,
        description="Initialize the accumulator values in adagrad. "
        "Mirrors initial_accumulator_value flag from tensorflow "
        "adagrad implementation (default 0.1 there).",
    )
    adam_beta1: float = Field(
        default=0.9,
        description="Beta1 parameter used by Adam. "
        "Almost without exception a value of 0.9 is used in the literature, "
        "seemingly giving good results, so we would discourage changing "
        "this value from the default without due consideration.",
    )
    adam_beta2: float = Field(
        default=0.999,
        description="Beta2 parameter used by Adam. "
        "Typically a value of 0.999 is recommended, "
        "as this is the value suggested by the original paper describing Adam, "
        "and is also the value adopted in other frameworks such as Tensorflow "
        "(https://www.tensorflow.org/api_docs/python/tf/train/AdamOptimizer) "
        "and Keras (https://keras.io/optimizers/). "
        "Whereas recently the paper Attention is All You Need suggested "
        "a value of 0.98 for beta2, this parameter may not work well "
        "for normal models / default baselines.",
    )
    adam_eps: float = Field(default=1e-8, description="Adam epsilon to forward to torch Optimizer.")
    weight_decay: float = Field(default=0.0, description="Weight decay to forward to torch Optimizer.")
    use_amp: bool = Field(default=True, description="Use torch mixed precision when compute_dtype is 16-bit.")
    learning_rate: float = Field(
        default=1.0,
        description="Starting learning rate. " "Recommended settings: " "sgd=1, adagrad=0.1, adadelta=1, adam=0.001.",
    )
    learning_rate_decay: float = Field(
        default=0.5,
        description="Decay learning rate by this much if steps have gone past start_decay_steps.",
    )
    start_decay_steps: int = Field(
        default=50000,
        description="Start decaying every decay_steps after this many steps.",
    )
    decay_steps: int = Field(default=10000, description="Frequency for learning rate decay, in steps.")
    decay_method: Literal["noam", "noamwd", "cosine", "rsqrt", "none"] = Field(
        default="none", description="Custom decay method to use."
    )
    warmup_steps: int = Field(default=4000, description="Number of warmup steps for custom decay.")
    reset_optim: Literal["none", "all", "states", "keep_states"] = Field(
        default="none", description="Optimization resetter when using train_from."
    )  # note: this is validated against update_vocab in TrainingConfig


class TrainingConfig(
    RunningConfig,
    OptimizerConfig,
    LoRaConfig,
    QuantizeConfig,
):

    model_config = get_config_dict()
    model_config["arbitrary_types_allowed"] = True  # to allow torch.dtype

    # Init stuff
    param_init: float = Field(
        default=0.1,
        description="Support value for uniform distribution parameters initialization. "
        "Set to 0 not to use initialization.",
    )
    param_init_method: Literal["xavier_uniform", "uniform", "normal"] = Field(
        default="uniform", description="Parameter initialization method."
    )
    freeze_encoder: bool = Field(default=False, description="Freeze parameters in encoder.")
    freeze_decoder: bool = Field(default=False, description="Freeze parameters in decoder.")
    # Pretrained word vectors (might be deprecated)
    pre_word_vecs_enc: str | None = Field(
        default=None,
        description="If a valid path is specified, will load pretrained " "word embeddings on the encoder side.",
    )
    pre_word_vecs_dec: str | None = Field(
        default=None,
        description="If a valid path is specified, will load pretrained " "word embeddings on the decoder side.",
    )
    data_type: str | None = "text"  # not really useful for now
    bucket_size: int = Field(
        default=262144,
        description="A bucket is a buffer of bucket_size examples "
        "to pick from the various corpora. "
        "The dynamic iterator batches batch_size items "
        "from the bucket and shuffle them.",
    )
    bucket_size_init: int = Field(
        default=-1,
        description="Bucket size is initialized with this amount of examples " "(see bucket_size_increment).",
    )
    bucket_size_increment: int = Field(
        default=0,
        description="Bucket size incremented with this amount of examples " "at each new bucket (up to bucket_size).",
    )
    prefetch_factor: int = Field(
        default=200,
        description="Number of mini-batches loaded in advance to avoid "
        "the GPU waiting during processing of next bucket.",
    )
    save_format: Literal["pytorch", "safetensors"] = Field(
        default="pytorch", description="Format to save the model weights."
    )
    save_checkpoint_steps: int = Field(default=5000, description="Frequency of checkpoint saving (in steps).")
    keep_checkpoint: int = Field(default=-1, description="Number of checkpoints to retain. (-1 retains all)")
    train_from: str | None = Field(
        default=None,
        description="Pretrained model/checkpoint weights to continue training from.",
    )
    num_workers: int = Field(
        default=2, description="Number of workers for pytorch.DataLoader objects."
    )  # dataloading related stuff (move to dataloader config ?
    batch_size: int = Field(default=64, description="Maximum batch size for training.")
    batch_size_multiple: int = Field(default=1, description="Batch size multiple for token batches.")
    batch_type: Literal["sents", "tokens"] = Field(default="sents", description="Batch grouping for batch_size.")
    normalization: Literal["sents", "tokens"] = Field(
        default="sents", description="Normalization method of the gradient."
    )
    accum_count: List[int] = Field(
        default=[1],
        description="Accumulate gradient this many times. "
        "Approximately equivalent to updating "
        "batch_size * accum_count batches at once. "
        "Recommended for transformer.",
    )
    accum_steps: List[int] = Field(default=[0], description="Steps at which accum_count values change.")
    valid_steps: int = Field(default=10000, description="Frequency of validation, in steps.")
    valid_batch_size: int = Field(default=32, description="Maximum batch size for validation.")
    train_steps: int = Field(default=100000, description="Number of training steps.")
    single_pass: bool = Field(
        default=False, description="Make a single pass over the training dataset."
    )  # deprecated? (not really manageable with weighted datasets)
    early_stopping: int = Field(
        default=0,
        description="Number of validation steps without improving that " "will trigger early stop of training.",
    )
    early_stopping_criteria: str | None = Field(
        default=None, description="Criteria to use for early stopping."
    )  # allowed values are not clear
    max_grad_norm: float = Field(
        default=5,
        description="If the norm of the gradient vector exceeds this value, "
        "renormalize it to have the norm equal to max_grad_norm.",
    )
    dropout: List[float] = Field(default=[0.3], description="Dropout probability.")
    attention_dropout: List[float] = Field(default=[0.1], description="Attention dropout probability.")
    dropout_steps: List[int] = Field(default=[0], description="Steps at which dropout changes.")
    label_smoothing: float = Field(
        default=0.0,
        description="Label smoothing value epsilon. "
        "Probability of all non-true labels will be "
        "smoothed by epsilon/(vocab_size-1). "
        "Set to 0 to turn off label smoothing. (https://arxiv.org/abs/1512.00567)",
    )
    average_decay: float = Field(
        default=0.0,
        description="Exponential moving average decay "
        "(https://en.wikipedia.org/wiki/Moving_average). "
        "Set to other than 0 (e.g. 1e-4) to activate. "
        "Similar to Marian NMT implementation "
        "(http://www.aclweb.org/anthology/P18-4020).",
    )
    average_every: int = Field(
        default=1,
        description="Step for moving average. Default is every update if average_decay is set.",
    )

    zero_out_prompt_loss: bool = Field(
        default=False,
        description="Set the prompt loss to zero. Mostly for LLM finetuning. "
        "Will be enabled only if the `insert_mask_before_placeholder` "
        "transform is applied.",
    )
    use_ckpting: List[str] = Field(
        default=[], description="Use gradient checkpointing for those modules."
    )  # validate on ["ffn", "mha", "lora"]
    update_vocab: bool = Field(
        default=False, description="Update source and target existing vocabularies."
    )  # not linked to model per se, rather to the training instance of the model

    lm_prior_model: str | None = Field(default=None, description="LM model to use to train the TM.")
    lm_prior_lambda: float = Field(default=0.0, description="LM Prior Lambda")
    lm_prior_tau: float = Field(default=1.0, description="LM Prior Tau")
    estim_loss_lambda: List[float] = Field(default=[1.0], description="Weight applied to estimator loss")
    estim_loss_lambda_steps: List[int] = Field(default=[0], description="Steps at which estimator loss lambda changes")
    score_threshold: float = Field(default=0.68, description="Threshold to filterout data")

    # Attention entropy configuration
    log_attention_entropy: bool = Field(
        default=True, description="Whether to compute and log attention entropy during training."
    )
    attention_entropy_types: List[str] | None = Field(
        default=None,
        description="Which attention types to compute entropy for. "
        "If None, computes for all available types (e.g., ['std', 'self', 'context']).",
    )
    attention_entropy_layers: List[int] | None = Field(
        default=None,
        description="Which attention layer indices to include in entropy computation. " "If None, includes all layers.",
    )
    attention_entropy_aggregation: Literal["mean", "max", "min"] = Field(
        default="mean", description="How to aggregate attention entropy across different attention types/layers."
    )

    @computed_field
    @cached_property
    def storage_dtype(self) -> torch.dtype:
        """
        Deduce which dtype to use for main model parameters.
        E.g. with mixed precision a copy is kept in float32.
        """
        if self.compute_dtype in [torch.float16, torch.bfloat16] and not self.use_amp:
            dtype = self.compute_dtype
            logger.info("Pure low precision training")
            logger.info("Non quantized layer compute is %s", self.compute_dtype)
            logger.info("This config has been tested to work only with bf16 / Adam(w)")
        else:
            dtype = torch.float32
            if self.compute_dtype in [torch.float16, torch.bfloat16]:
                logger.info("Model weights will be float32 (torch AMP)")
                logger.info("Non quantized layer compute is %s", self.compute_dtype)
        return dtype

    @field_validator("use_ckpting", mode="after")
    @classmethod
    def checkpointing_layers(cls, field: List[str]):
        _layers = {"ffn", "mha", "lora"}
        if set(field) | _layers != _layers:
            raise ValueError(f"use_ckpting only supports these layers: {_layers}")
        return field

    def get_model_path(self):
        # TODO: clarify this
        if self.train_from is not None:
            return self.train_from
        else:
            return self.model_path

    @model_validator(mode="after")
    def _validate_running_config(self):
        self.check_self_attn_backend()
        # self._validate_language_model_compatibilities_opts()
        if self.world_size < len(self.gpu_ranks):
            raise AssertionError("parameter counts of -gpu_ranks must be less or equal " "than -world_size.")
        if self.world_size == len(self.gpu_ranks) and self.world_size > 0 and min(self.gpu_ranks) > 0:
            raise AssertionError(
                "-gpu_ranks should have master(=0) rank " "unless -world_size is greater than len(gpu_ranks)."
            )

        assert len(self.dropout) == len(self.dropout_steps), "Number of dropout values must match accum_steps values"

        assert len(self.attention_dropout) == len(
            self.dropout_steps
        ), "Number of attention_dropout values must match accum_steps values"

        assert len(self.accum_count) == len(
            self.accum_steps
        ), "Number of accum_count values must match number of accum_steps"

        if self.update_vocab:
            assert self.train_from, "-update_vocab needs -train_from option"
            assert self.reset_optim in [
                "states",
                "all",
            ], '-update_vocab needs -reset_optim "states" or "all"'

        if (
            "param_init" in self.model_fields_set
            and self.param_init != 0.0
            and self.param_init_method == "xavier_uniform"
        ):
            logger.warn(f"xavier_uniform initialization does not require param_init ({self.param_init})")

        assert (
            self.compute_dtype != torch.int8
        ), "int8 compute_dtype is currently only used for inference dynamic quantization"

        return self



import copy
import unittest
import torch
import pyonmttok
from eole.constants import DefaultTokens
from collections import Counter
import eole
import eole.inputters

from eole.models.model import build_src_emb, build_tgt_emb, build_encoder, build_decoder
from eole.config.run import TrainConfig
from eole.config.data import Dataset
from eole.config.models import CustomModelConfig

# In theory we should probably call ModelConfig here,
# but we can't because model building relies on some params
# not currently in model config (dropout, freeze_word_vecs_enc, etc.)


class NoOpPosition:
    """A no-op position encoding callable."""

    def update(self, *args, **kwargs):
        return None


opt = TrainConfig(
    data={"dummy": Dataset(path_src="eole/tests/data/src-train.txt")},  # actual file path (tested in validate)
    src_vocab="dummy",
    share_vocab=True,
    model=CustomModelConfig(),
)  # now required by validation


class TestModel(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super(TestModel, self).__init__(*args, **kwargs)
        self.opt = opt
        self.opt.training.self_attn_backend = "pytorch"
        self.rope = NoOpPosition()

    def get_vocabs(self):
        src_vocab = pyonmttok.build_vocab_from_tokens(
            Counter(),
            maximum_size=0,
            minimum_frequency=1,
            special_tokens=[
                DefaultTokens.UNK,
                DefaultTokens.PAD,
                DefaultTokens.BOS,
                DefaultTokens.EOS,
            ],
        )

        tgt_vocab = pyonmttok.build_vocab_from_tokens(
            Counter(),
            maximum_size=0,
            minimum_frequency=1,
            special_tokens=[
                DefaultTokens.UNK,
                DefaultTokens.PAD,
                DefaultTokens.BOS,
                DefaultTokens.EOS,
            ],
        )

        vocabs = {
            "src": src_vocab,
            "tgt": tgt_vocab,
            "specials": {
                "bos_token": DefaultTokens.BOS,
                "pad_token": DefaultTokens.PAD,
                "eos_token": DefaultTokens.EOS,
                "unk_token": DefaultTokens.UNK,
            },
        }
        return vocabs

    def get_batch(self, source_l=3, bsize=1):
        # batch x len
        test_src = torch.ones(bsize, source_l).long()
        test_tgt = torch.ones(bsize, source_l).long()
        test_length = torch.ones(bsize).fill_(source_l).long()
        return test_src, test_tgt, test_length

    def embeddings_forward(self, opt, source_l=3, bsize=1):
        """
        Tests if the embeddings works as expected

        args:
            opt: set of options
            source_l: Length of generated input sentence
            bsize: Batchsize of generated input
        """
        vocabs = self.get_vocabs()
        emb = build_src_emb(opt.model, vocabs, running_config=opt.training)
        test_src, _, __ = self.get_batch(source_l=source_l, bsize=bsize)
        if opt.model.decoder.decoder_type == "transformer":
            input = torch.cat([test_src, test_src], 1)
            res = emb(input)
            compare_to = torch.zeros(bsize, source_l * 2, opt.model.embeddings.src_word_vec_size)
        else:
            res = emb(test_src)
            compare_to = torch.zeros(bsize, source_l, opt.model.embeddings.src_word_vec_size)

        self.assertEqual(res.size(), compare_to.size())

    def encoder_forward(self, opt, source_l=3, bsize=1):
        """
        Tests if the encoder works as expected

        args:
            opt: set of options
            source_l: Length of generated input sentence
            bsize: Batchsize of generated input
        """
        # if opt.hidden_size > 0:
        #     opt.encoder.hidden_size = opt.hidden_size
        vocabs = self.get_vocabs()
        embeddings = build_src_emb(opt.model, vocabs, running_config=opt.training)
        enc = build_encoder(opt.model, running_config=opt.training)

        test_src, test_tgt, test_length = self.get_batch(source_l=source_l, bsize=bsize)
        emb_src = embeddings(test_src)
        pad_idx = embeddings.word_padding_idx
        pad_mask = test_src.eq(pad_idx).unsqueeze(1)
        enc_out, hidden_t = enc(emb_src, pad_mask=pad_mask)

        # Initialize vectors to compare size with
        test_hid = torch.zeros(opt.model.encoder.layers, bsize, opt.model.encoder.hidden_size)
        test_out = torch.zeros(bsize, source_l, opt.model.decoder.hidden_size)

        # Ensure correct sizes and types
        self.assertEqual(test_hid.size(), hidden_t[0].size(), hidden_t[1].size())
        self.assertEqual(test_out.size(), enc_out.size())
        self.assertEqual(type(enc_out), torch.Tensor)

    def model_forward(self, opt, source_l=3, bsize=1):
        """
        Creates a model with a custom opt function.
        Forwards a testbatch and checks output size.

        Args:
            opt: Namespace with options
            source_l: length of input sequence
            bsize: batchsize
        """
        # should be already done in validator
        # if opt.hidden_size > 0:
        #     opt.encoder.hidden_size = opt.hidden_size
        #     opt.decoder.hidden_size = opt.hidden_size
        vocabs = self.get_vocabs()

        src_emb = build_src_emb(opt.model, vocabs, running_config=opt.training)
        enc = build_encoder(opt.model, running_config=opt.training)

        tgt_emb = build_tgt_emb(
            opt.model,
            vocabs,
            running_config=opt.training,
        )
        dec = build_decoder(opt.model, running_config=opt.training)

        model = eole.models.model.EncoderDecoderModel(
            encoder=enc,
            decoder=dec,
            src_emb=src_emb,
            tgt_emb=tgt_emb,
            hidden_size=opt.model.decoder.hidden_size,
            rope=NoOpPosition(),
        )
        test_src, test_tgt, test_length = self.get_batch(source_l=source_l, bsize=bsize)
        output, attn, estim = model(test_src, test_tgt, test_length)
        outputsize = torch.zeros(bsize, source_l - 1, opt.model.decoder.hidden_size)
        # Make sure that output has the correct size and type
        self.assertEqual(output.size(), outputsize.size())
        self.assertEqual(type(output), torch.Tensor)


def _add_test(param_setting, methodname):
    """
    Adds a Test to TestModel according to settings

    Args:
        param_setting: list of tuples of (param, setting)
        methodname: name of the method that gets called
    """

    def test_method(self):
        opt = copy.deepcopy(self.opt)
        if param_setting:
            for param, setting in param_setting:
                setattr(opt, param, setting)
        getattr(self, methodname)(opt)

    if param_setting:
        name = "test_" + methodname + "_" + "_".join(str(param_setting).split())
    else:
        name = "test_" + methodname + "_standard"
    setattr(TestModel, name, test_method)
    test_method.__name__ = name


"""
TEST PARAMETERS
"""
# opt.brnn = False # deprecated and not used here

test_embeddings = [
    [("model", {"architecture": "rnn"})],
    [("model", {"architecture": "transformer"})],
]

for p in test_embeddings:
    _add_test(p, "embeddings_forward")

tests_encoder = [
    [
        # ("encoder_type", "mean"),
        (
            "model",
            {
                "architecture": "custom",
                "encoder": {"encoder_type": "mean"},
                "decoder": {"decoder_type": "rnn"},
            },
        )
    ],
    # [('encoder_type', 'transformer'),
    # ('word_vec_size', 16), ('hidden_size', 16)],
]

for p in tests_encoder:
    _add_test(p, "encoder_forward")

tests_model = [
    # [("rnn_type", "GRU")],
    [
        (
            "model",
            {
                "architecture": "rnn",
                "encoder": {"rnn_type": "GRU"},
                "decoder": {"rnn_type": "GRU"},
            },
        )
    ],
    [("model", {"architecture": "rnn", "layers": 10})],
    [("model", {"architecture": "rnn", "input_feed": 0})],
    [
        (
            "model",
            {
                "architecture": "transformer",
                "hidden_size": 16,
                "embeddings": {"word_vec_size": 16},
                # "word_vec_size": 16,  # can not be set at model level right now
                # (would need to define field in ModelConfig)
            },
        )
    ],  # not sure what this one is for (and how it works without position_encoding)
    [
        (
            "model",
            {
                "architecture": "transformer",
                "hidden_size": 16,
                "embeddings": {"word_vec_size": 16, "position_encoding": True},
            },
        )
        # ("decoder_type", "transformer"),
        # ("encoder_type", "transformer"),
        # ("src_word_vec_size", 16),
        # ("tgt_word_vec_size", 16),
        # ("position_encoding", True),
    ],
    [
        ("model", {"architecture": "rnn", "decoder": {"coverage_attn": True}})
        # ideally we should not have to pass decoder_type to default to rnn,
        # but pydantic does not seem to agree
    ],
    [("model", {"architecture": "rnn", "decoder": {}})],
    [
        ("model", {"architecture": "rnn", "decoder": {"global_attention": "mlp"}})
        # ("global_attention", "mlp"),
    ],
    [
        ("model", {"architecture": "rnn", "decoder": {"context_gate": "both"}})
        # ("context_gate", "both"),
    ],
    [
        (("model", {"architecture": "rnn", "decoder": {"context_gate": "target"}}))
        # ("context_gate", "target"),
    ],
    [("model", {"architecture": "rnn", "decoder": {"context_gate": "source"}})],
    # [("encoder_type", "brnn"), ("brnn_merge", "sum")],
    # not sure about this one, brnn_merge does not seem to exist in the codebase
    [
        ("model", {"architecture": "rnn", "encoder": {"encoder_type": "brnn"}})
        # ("encoder_type", "brnn"),
    ],
    [
        (
            "model",
            {
                "architecture": "cnn",
            },
        )
        # ("decoder_type", "cnn"), ("encoder_type", "cnn"),
    ],
    [
        ("model", {"architecture": "rnn", "decoder": {"global_attention": None}})
        # ("encoder_type", "rnn"), ("global_attention", None),
    ],
    [
        (
            "model",
            {
                "architecture": "rnn",
                "decoder": {
                    "global_attention": None,
                },
            },
        )
        # ("encoder_type", "rnn"),
        # ("global_attention", None),
    ],
    [
        (
            "model",
            {
                "architecture": "rnn",
                "decoder": {
                    "decoder_type": "rnn",  # not necessary, but adding here to test the case
                    "global_attention": "mlp",
                },
            },
        )
        # ("encoder_type", "rnn"),
        # ("global_attention", "mlp"),
    ],
]


for p in tests_model:
    _add_test(p, "model_forward")



#!/usr/bin/env python
from setuptools import setup, find_packages
from os import path

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="eole",
    description="Open language modeling toolkit based on PyTorch",
    long_description=long_description,
    long_description_content_type="text/markdown",
    version="0.4.1",
    packages=find_packages(),
    project_urls={
        "Source": "https://github.com/eole-nlp/eole/",
    },
    python_requires=">=3.10",
    install_requires=[
        "configargparse",
        "ctranslate2>=4,<5",
        "fastapi",
        "fasttext-wheel",
        "huggingface_hub",
        "datasets",
        "numpy>=2.0",
        "pandas",
        "protobuf==3.20.1",
        "pyahocorasick",
        "pyonmttok>=1.37,<2",
        "pyyaml",
        "rapidfuzz",
        "rich",
        "sacrebleu",
        "safetensors",
        "sentencepiece>=0.1.94,<0.1.98",
        "six",
        "spacy",
        "subword-nmt>=0.3.7",
        "tensorboard>=2.18.0",
        "torch>=2.8,<2.10",
        "torch-optimi",
        "uvicorn",
        "waitress",
        "pydantic",
    ],
    entry_points={
        "console_scripts": [
            "eole=eole.bin.main:main",
        ],
    },
)



"""Statistics calculation utility"""

import time
import math
import sys

from eole.utils.logging import logger


class Statistics(object):
    """
    Accumulator for loss statistics.
    Currently calculates:

    * accuracy
    * perplexity
    * elapsed time
    """

    def __init__(
        self,
        loss=0,
        auxloss=0,
        n_batchs=0,
        n_sents=0,
        n_tokens=0,
        n_correct=0,
        computed_metrics=None,
        data_stats=None,
        attention_entropy=0,
        n_attention_samples=0,
    ):
        self.loss = loss
        self.auxloss = auxloss
        self.n_batchs = n_batchs
        self.n_sents = n_sents
        self.n_tokens = n_tokens
        self.n_correct = n_correct
        self.n_src_tokens = 0
        self.computed_metrics = computed_metrics if computed_metrics is not None else {}
        self.data_stats = data_stats if data_stats is not None else {}
        self.attention_entropy = attention_entropy
        self.n_attention_samples = n_attention_samples
        self.start_time = time.time()

    @staticmethod
    def all_gather_stats(stat, max_size=4096):
        """
        Gather a `Statistics` object accross multiple process/nodes

        Args:
            stat(:obj:Statistics): the statistics object to gather
                accross all processes/nodes
            max_size(int): max buffer size to use

        Returns:
            `Statistics`, the update stats object
        """
        stats = Statistics.all_gather_stats_list([stat], max_size=max_size)
        return stats[0]

    @staticmethod
    def all_gather_stats_list(stat_list, max_size=4096):
        """
        Gather a `Statistics` list accross all processes/nodes

        Args:
            stat_list(list([`Statistics`])): list of statistics objects to
                gather accross all processes/nodes
            max_size(int): max buffer size to use

        Returns:
            our_stats(list([`Statistics`])): list of updated stats
        """
        from torch.distributed import get_rank
        from eole.utils.distributed import all_gather_list

        # Get a list of world_size lists with len(stat_list) Statistics objects
        all_stats = all_gather_list(stat_list, max_size=max_size)

        our_rank = get_rank()
        our_stats = all_stats[our_rank]
        for other_rank, stats in enumerate(all_stats):
            if other_rank == our_rank:
                continue
            for i, stat in enumerate(stats):
                our_stats[i].update(stat, update_n_src_tokens=True)
        return our_stats

    def update(self, stat, update_n_src_tokens=False):
        """
        Update statistics by suming values with another `Statistics` object

        Args:
            stat: another statistic object
            update_n_src_tokens(bool): whether to update (sum) `n_src_tokens`
                or not

        """
        self.loss += stat.loss
        self.auxloss += stat.auxloss
        self.n_batchs += stat.n_batchs
        self.n_sents += stat.n_sents
        self.n_tokens += stat.n_tokens
        self.n_correct += stat.n_correct
        self.attention_entropy += stat.attention_entropy
        self.n_attention_samples += stat.n_attention_samples
        self.computed_metrics = stat.computed_metrics
        for cid in stat.data_stats.keys():
            if cid in self.data_stats.keys():
                self.data_stats[cid]["count"] += stat.data_stats[cid]["count"]
            else:
                self.data_stats[cid] = {}
                self.data_stats[cid]["count"] = stat.data_stats[cid]["count"]
            self.data_stats[cid]["index"] = stat.data_stats[cid]["index"]

        if update_n_src_tokens:
            self.n_src_tokens += stat.n_src_tokens

    def computed_metric(self, metric):
        """check if metric(TER/BLEU) is computed and return it"""
        assert metric in self.computed_metrics, "Metric {} not found".format(metric)
        return self.computed_metrics[metric]

    def accuracy(self):
        """compute accuracy"""
        return 100 * (self.n_correct / self.n_tokens)

    def xent(self):
        """compute cross entropy"""
        return self.loss / self.n_tokens

    def aux_loss(self):
        return self.auxloss / self.n_sents

    def ppl(self):
        """compute perplexity"""
        return math.exp(min(self.loss / self.n_tokens, 100))

    def avg_attention_entropy(self):
        """compute average attention entropy"""
        if self.n_attention_samples > 0:
            return self.attention_entropy / self.n_attention_samples
        return 0.0

    def elapsed_time(self):
        """compute elapsed time"""
        return time.time() - self.start_time

    def output(self, step, num_steps, learning_rate, start):
        """Write out statistics to stdout.

        Args:
           step (int): current step
           n_batch (int): total batches
           start (int): start time of step.
        """
        t = self.elapsed_time()
        step_fmt = "%2d" % step
        if num_steps > 0:
            step_fmt = "%s/%5d" % (step_fmt, num_steps)
        logger.info(
            (
                "Step %s; acc: %2.1f; ppl: %5.2f; xent: %2.2f; aux: %2.3f; "
                + "attn_ent: %2.3f; lr: %7.2e; sents: %7.0f; bsz: %4.0f/%4.0f/%2.0f; "
                + "%3.0f/%3.0f tok/s; %6.0f sec;"
            )
            % (
                step_fmt,
                self.accuracy(),
                self.ppl(),
                self.xent(),
                self.aux_loss(),
                self.avg_attention_entropy(),
                learning_rate,
                self.n_sents,
                self.n_src_tokens / self.n_batchs,
                self.n_tokens / self.n_batchs,
                self.n_sents / self.n_batchs,
                self.n_src_tokens / (t + 1e-5),
                self.n_tokens / (t + 1e-5),
                time.time() - start,
            )
            + "".join([" {}: {}".format(k, round(v, 2)) for k, v in self.computed_metrics.items()])
        )
        sys.stdout.flush()

    def log_tensorboard(self, prefix, writer, learning_rate, patience, step):
        """display statistics to tensorboard"""
        t = self.elapsed_time()
        writer.add_scalar(prefix + "/xent", self.xent(), step)
        writer.add_scalar(prefix + "/ppl", self.ppl(), step)
        writer.add_scalar(prefix + "/attention_entropy", self.avg_attention_entropy(), step)
        for k, v in self.computed_metrics.items():
            writer.add_scalar(prefix + "/" + k, round(v, 4), step)
        writer.add_scalar(prefix + "/accuracy", self.accuracy(), step)
        writer.add_scalar(prefix + "/tgtper", self.n_tokens / t, step)
        writer.add_scalar(prefix + "/lr", learning_rate, step)
        if patience is not None:
            writer.add_scalar(prefix + "/patience", patience, step)



"""
Implementation of "Convolutional Sequence to Sequence Learning"
"""

import torch.nn as nn

from eole.encoders.encoder import EncoderBase
from eole.utils.cnn_factory import shape_transform, StackedCNN

SCALE_WEIGHT = 0.5**0.5


class CNNEncoder(EncoderBase):
    """Encoder based on "Convolutional Sequence to Sequence Learning"
    :cite:`DBLP:journals/corr/GehringAGYD17`.
    """

    def __init__(self, model_config, running_config=None):
        super(CNNEncoder, self).__init__()

        input_size = model_config.hidden_size  # we need embeddings.src_word_vec_size instead
        self.linear = nn.Linear(input_size, model_config.hidden_size)
        self.cnn = StackedCNN(
            model_config.layers,
            model_config.hidden_size,
            model_config.cnn_kernel_width,
            getattr(running_config, "dropout", [0.0])[0],
        )

    @classmethod
    def from_config(cls, model_config, running_config=None):
        """Alternate constructor."""
        # config = opt.model.encoder  # CnnEncoderConfig
        return cls(
            model_config,  # CnnEncoderConfig
            running_config,  # might be better to set this out of the model building logic (update_dropout call sometime in training) # noqa: E501
        )

    def forward(self, emb, **kwargs):
        """See :func:`EncoderBase.forward()`"""
        # batch x len x dim

        emb_reshape = emb.view(emb.size(0) * emb.size(1), -1)
        emb_remap = self.linear(emb_reshape)
        emb_remap = emb_remap.view(emb.size(0), emb.size(1), -1)
        emb_remap = shape_transform(emb_remap)
        out = self.cnn(emb_remap)

        return out.squeeze(3), emb_remap.squeeze(3)

    def update_dropout(self, dropout, attention_dropout=None):
        self.cnn.dropout.p = dropout



"""
Base class for vision encoders.
We'll start with Pixtral encoder, but we can build up on it later.
https://github.com/mistralai/mistral-inference/pull/217/files#diff-5cf8b83293298d0bdda514a77cae34565cb54571cddab02801d99a52f0f71a66
"""

import torch
import torch.nn as nn
import torch.nn.functional as F
import math
from typing import Optional

from eole.encoders.transformer import TransformerEncoderLayer
from eole.modules.rope import build_rope
from eole.constants import PositionEncodingType, LayerNorm


class PatchMerger(nn.Module):
    """
    Learned merging of spatial_merge_size ** 2 patches
    """

    def __init__(self, model_config):
        super().__init__()
        self.config = model_config
        hidden_size = model_config.encoder.hidden_size
        self.spatial_merge_size = model_config.spatial_merge_size
        self.patch_size = model_config.encoder.patch_size
        self.merging_layer = nn.Linear(hidden_size * self.spatial_merge_size**2, hidden_size, bias=False)

    def forward(self, image_features: torch.Tensor, image_sizes: torch.Tensor) -> torch.Tensor:
        image_sizes = [
            (image_size[0] // self.patch_size, image_size[1] // self.patch_size) for image_size in image_sizes
        ]
        tokens_per_image = [h * w for h, w in image_sizes]
        d = image_features.shape[-1]

        flattened_features = image_features.view(-1, d)
        permuted_tensor = []
        for image_index, image_tokens in enumerate(flattened_features.split(tokens_per_image)):
            # Reshape image_tokens into a 2D grid
            h, w = image_sizes[image_index]
            image_grid = image_tokens.view(h, w, d).permute(2, 0, 1).unsqueeze(0)
            grid = torch.nn.functional.unfold(
                image_grid, kernel_size=self.spatial_merge_size, stride=self.spatial_merge_size
            )
            grid = grid.view(d * self.spatial_merge_size**2, -1).t()
            permuted_tensor.append(grid)

        image_features = torch.cat(permuted_tensor, dim=0)
        image_features = self.merging_layer(image_features.unsqueeze(0))

        return image_features


def position_ids_in_meshgrid(patch_embeds_list, max_width):
    positions = []
    for patch in patch_embeds_list:
        height, width = patch.shape[-2:]
        mesh = torch.meshgrid(torch.arange(height), torch.arange(width), indexing="ij")
        h_grid, v_grid = torch.stack(mesh, dim=-1).reshape(-1, 2).chunk(2, -1)
        ids = h_grid * max_width + v_grid
        positions.append(ids[:, 0])
    return positions


def create_block_diagonal_mask(lengths, device):
    """
    Create a block diagonal mask based on sequence lengths.
    Args:
        lengths (list of int): List of lengths for each block.
    Returns:
        torch.Tensor: A 2D boolean tensor with True in the
            block diagonal regions and False elsewhere.
    """
    total_length = sum(lengths)
    mask = torch.zeros((total_length, total_length), dtype=torch.bool)

    start = 0
    for length in lengths:
        end = start + length
        mask[start:end, start:end] = True
        start = end

    return mask.to(device)


def get_abs_pos(abs_pos, tgt_size):
    dim = abs_pos.size(-1)
    abs_pos_new = abs_pos.squeeze(0)
    cls_token, old_pos_embed = abs_pos_new[:1], abs_pos_new[1:]

    src_size = int(math.sqrt(abs_pos_new.shape[0] - 1))
    tgt_size = int(math.sqrt(tgt_size))
    dtype = abs_pos.dtype

    if src_size != tgt_size:
        old_pos_embed = old_pos_embed.view(1, src_size, src_size, dim).permute(0, 3, 1, 2).contiguous()
        old_pos_embed = old_pos_embed.to(torch.float32)
        new_pos_embed = F.interpolate(
            old_pos_embed,
            size=(tgt_size, tgt_size),
            mode="bicubic",
            antialias=True,
            align_corners=False,
        ).to(dtype)
        new_pos_embed = new_pos_embed.permute(0, 2, 3, 1)
        new_pos_embed = new_pos_embed.view(tgt_size * tgt_size, dim)
        vision_pos_embed = torch.cat([cls_token, new_pos_embed], dim=0)
        vision_pos_embed = vision_pos_embed.view(1, tgt_size * tgt_size + 1, dim)
        return vision_pos_embed
    else:
        return abs_pos


# Clip() based vision encoder
class VisionEncoder(nn.Module):
    def __init__(self, encoder_config, running_config=None):
        super(VisionEncoder, self).__init__()
        self.encoder_config = encoder_config
        if encoder_config.encoder_sam:
            from eole.encoders.deepseek_sam import build_sam_vit_b

            self.sam = build_sam_vit_b()
        else:
            self.sam = None
        if encoder_config.position_encoding_type == PositionEncodingType.Learned:
            n_positions = encoder_config.n_positions
            self.interpolate_mode = encoder_config.interpolate_mode
            if self.interpolate_mode is not None:
                self.max_patch_per_side = int(n_positions**0.5)
                n_positions += 1
            self.position_embeddings = nn.Embedding(
                n_positions,
                encoder_config.hidden_size,
            )
            self.register_buffer("position_ids", torch.arange(n_positions).expand((1, -1)))
        else:
            self.rope = build_rope(encoder_config, mode="2d")
        self.patch_conv = nn.Conv2d(
            in_channels=encoder_config.num_channels,
            out_channels=encoder_config.hidden_size,
            kernel_size=encoder_config.patch_size,
            stride=encoder_config.patch_size,
            bias=encoder_config.patch_conv_bias,
        )
        if encoder_config.use_class_embedding:
            self.class_embedding = torch.nn.Embedding(1, encoder_config.hidden_size)
        else:
            self.class_embedding = None
        if encoder_config.layernorm_pre:
            self.ln_pre = LayerNorm[encoder_config.layer_norm](encoder_config.hidden_size, eps=1e-5)
        else:
            self.ln_pre = None
        self.transformer_layers = torch.nn.ModuleList()
        for _ in range(encoder_config.layers):
            self.transformer_layers.append(TransformerEncoderLayer(encoder_config, running_config=running_config))

        if encoder_config.layernorm_post:
            self.post_layernorm = nn.LayerNorm(encoder_config.hidden_size, eps=encoder_config.norm_eps)
        else:
            self.post_layernorm = None

        head_dim = encoder_config.hidden_size // encoder_config.heads
        assert head_dim % 2 == 0, "ROPE requires even head_dim"
        self._freqs_cis: Optional[torch.Tensor] = None

    @classmethod
    def from_config(cls, encoder_config, running_config=None):
        """Alternate constructor."""
        return cls(
            encoder_config,
            running_config,
        )

    @property
    def device(self):
        return next(self.parameters()).device

    def forward(self, images, sam_patches=None):
        """
        Args:
            images: list of N_img images of variable sizes, each of shape (C, H, W)
        Returns:
            image_features: tensor of token features for all tokens of all images of
                shape (N_img, Seqlen, D)
        """

        # TODO add as @property somewhere
        dtype = next(self.parameters()).dtype
        mask = None
        position_embeddings = None

        if sam_patches is not None:
            patch_embeds = sam_patches
            # flatten H+W then change to (H+W, C)
            patch_embeds = patch_embeds.flatten(2).transpose(1, 2)
            if self.class_embedding is not None:  # deepseekocr
                bs = patch_embeds.size(0)
                class_embeds = self.class_embedding.weight.expand(bs, -1, -1)
                patch_embeds = torch.cat([class_embeds, patch_embeds], dim=1)
                patch_embeds += get_abs_pos(self.position_embeddings(self.position_ids), patch_embeds.size(1))
            if self.ln_pre is not None:
                patch_embeds = self.ln_pre(patch_embeds)
        else:
            patch_embeds_list = [self.patch_conv(img.to(dtype)) for img in images]
            positions = position_ids_in_meshgrid(
                patch_embeds_list,
                max_width=self.encoder_config.image_size // self.encoder_config.patch_size,
            )
            if hasattr(self, "position_embeddings"):
                # Learned positions (Gemma3, Classic Clip)
                patch_embeds = torch.cat([p.unsqueeze(0) for p in patch_embeds_list], dim=0)
                B, D, H, W = patch_embeds.size()
                patch_embeds = patch_embeds.flatten(2).transpose(1, 2)

                if self.interpolate_mode is None:  # Gemma3 / DeepseekOCR
                    patch_pos_embed = self.position_embeddings(torch.stack(positions).to(self.device))
                else:  # HunyuanOCR
                    patch_pos_shape = (1, self.max_patch_per_side, self.max_patch_per_side, D)
                    patch_pos_embed = (
                        self.position_embeddings.weight[1:, :].reshape(patch_pos_shape).permute(0, 3, 1, 2).float()
                    )
                    h0 = H + 0.1  # taken from the HunyuanPCR implementation
                    w0 = W + 0.1
                    patch_pos_embed = nn.functional.interpolate(
                        patch_pos_embed,
                        scale_factor=((h0 / self.max_patch_per_side), (w0 / self.max_patch_per_side)),
                        mode=self.interpolate_mode,
                        align_corners=False,
                    )
                    patch_pos_embed = patch_pos_embed.reshape(D, -1).transpose(0, 1).unsqueeze(0).to(patch_embeds.dtype)

                patch_embeds += patch_pos_embed
            else:  # rope embedding / pixtral / Mistral
                patch_embeds = torch.cat([p.flatten(1).permute(1, 0) for p in patch_embeds_list], dim=0)
                if self.ln_pre is not None:
                    patch_embeds = self.ln_pre(patch_embeds)
                # create a fake batch dim
                patch_embeds = patch_embeds.unsqueeze(0)
                mask = create_block_diagonal_mask(
                    [p.shape[-2] * p.shape[-1] for p in patch_embeds_list],
                    device=self.device,
                )  # block diagonal mask delimiting images
                # revert for proper downstream compatibility
                mask = ~mask
                position_embeddings = self.rope.update(
                    patch_embeds.size(1),
                    step=0,
                    reset=True,
                    positions=torch.cat(positions).to(self.device),
                )

        out = patch_embeds

        for i, layer in enumerate(self.transformer_layers):
            out = layer(out, pad_mask=mask, position_embeddings=position_embeddings)

        if self.post_layernorm is not None:
            out = self.post_layernorm(out)

        return out


# TODO refactor the 3 projectors below for unique code path
# Multi-Modal Projector
class VisionLanguageAdapter(nn.Module):
    def __init__(self, model_config):
        super(VisionLanguageAdapter, self).__init__()
        in_dim = model_config.encoder.hidden_size
        out_dim = model_config.decoder.hidden_size
        bias = getattr(model_config, "adapter_bias", False)
        self.has_patch = False
        if model_config.spatial_merge_size > 1:
            self.has_patch = True
            self.layernorm = LayerNorm[model_config.encoder.layer_norm](in_dim, eps=1e-5)
            self.patch_merger = PatchMerger(model_config)
        self.w_in = nn.Linear(in_dim, out_dim, bias=bias)
        self.gelu = nn.GELU()
        self.w_out = nn.Linear(out_dim, out_dim, bias=bias)

    def forward(self, x, image_sizes=None):
        if self.has_patch:
            x = self.layernorm(x)
            x = self.patch_merger(x, image_sizes)
        return self.w_out(self.gelu(self.w_in(x)))

    @classmethod
    def from_config(cls, model_config, running_config=None):
        return cls(
            model_config,
        )


class Gemma3MultiModalProjector(nn.Module):
    # https://github.com/huggingface/transformers/blob/071a161d3e38f56dbda2743b979f0afeed2cd4f1/src/transformers/models/gemma3/modular_gemma3.py#L717
    def __init__(self, in_dim, out_dim, image_size, patch_size, mm_tokens_per_image):
        super(Gemma3MultiModalProjector, self).__init__()
        self.w_in = nn.Linear(in_dim, out_dim, bias=False)
        self.norm = LayerNorm["gemma-rms"](in_dim)  # forced because no value in config file
        self.patches_per_image = int(image_size / patch_size)
        self.tokens_per_side = int(mm_tokens_per_image**0.5)
        self.kernel_size = self.patches_per_image // self.tokens_per_side
        self.avg_pool = nn.AvgPool2d(kernel_size=self.kernel_size, stride=self.kernel_size)

    def forward(self, x, image_sizes):
        batch_size, _, hidden_size = x.size()
        reshaped = (
            x.transpose(1, 2)
            .reshape(batch_size, hidden_size, self.patches_per_image, self.patches_per_image)
            .contiguous()
        )
        pooled = self.avg_pool(reshaped).flatten(2).transpose(1, 2)
        normed = self.norm(pooled)
        projected = self.w_in(normed)
        return projected.type_as(x)

    @classmethod
    def from_config(cls, model_config, running_config=None):
        return cls(
            in_dim=model_config.encoder.hidden_size,
            out_dim=model_config.decoder.hidden_size,
            image_size=model_config.encoder.image_size,
            patch_size=model_config.encoder.patch_size,
            mm_tokens_per_image=model_config.encoder.mm_tokens_per_image,
        )


# Multi-Modal Projector
class DeepSeekOCRProjector(nn.Module):
    def __init__(self, model_config):
        super(DeepSeekOCRProjector, self).__init__()
        bias = getattr(model_config, "adapter_bias", False)
        self.w_in = nn.Linear(model_config.encoder.hidden_size * 2, model_config.decoder.hidden_size, bias=bias)

    def forward(self, x, image_sizes=None):
        return self.w_in(x)

    @classmethod
    def from_config(cls, model_config, running_config=None):
        return cls(
            model_config,
        )


class HunYuanVisionPatchMerger(nn.Module):
    def __init__(
        self,
        model_config,
    ):
        super().__init__()
        in_channels = model_config.encoder.hidden_size
        out_channels = model_config.decoder.hidden_size
        rms_norm_eps = getattr(model_config.encoder, "norm_eps", 1e-5)
        self.patch_size = model_config.encoder.patch_size
        self.spatial_merge_size = model_config.spatial_merge_size
        embed_std = out_channels**-0.5

        self.proj = nn.Sequential(
            nn.Conv2d(
                in_channels,
                in_channels * 2,
                kernel_size=self.spatial_merge_size,
                stride=self.spatial_merge_size,
            ),
            nn.GELU(),
            nn.Conv2d(in_channels * 2, in_channels * 4, kernel_size=1),
        )
        self.mlp = nn.Linear(in_channels * 4, out_channels)

        self.image_newline = nn.Parameter(torch.randn(in_channels * 4) * embed_std)
        self.image_begin = nn.Parameter(torch.randn(out_channels) * embed_std)
        self.image_end = nn.Parameter(torch.randn(out_channels) * embed_std)
        self.image_sep = nn.Parameter(torch.randn(out_channels) * embed_std)

        self.before_rms = LayerNorm["rms"](in_channels, eps=rms_norm_eps)
        self.after_rms = LayerNorm["rms"](out_channels, eps=rms_norm_eps)

    def forward(self, x, image_sizes):
        """
        Args:
            x: Tensor of shape (batch_size, seq_len, hidden_size)
            image_sizes: Tensor of shape (batch_size, 2), where each entry is (width, height).
        Raises:
            ValueError: If images in the batch have different sizes.
        Note:
            This implementation requires all images in the batch to have the same dimensions.
        """
        # Validate that all images in the batch have the same dimensions
        if not torch.all(image_sizes == image_sizes[0]):
            raise ValueError(
                "All images in the batch must have the same dimensions. " f"Got image_sizes: {image_sizes.tolist()}"
            )
        x = self.before_rms(x)
        h, w = image_sizes[0][1].item() // self.patch_size, image_sizes[0][0].item() // self.patch_size
        dtype = x.dtype
        x = x.permute(0, 2, 1).reshape(x.shape[0], -1, h, w)
        x = self.proj(x)  # b,c,h,w
        b, c, h, w = x.shape
        x = torch.cat(
            [x, self.image_newline.reshape(1, c, 1, 1).expand(b, c, h, 1).to(dtype)],
            dim=-1,
        )
        x = x.reshape(b, c, -1).permute(0, 2, 1)
        x = self.mlp(x)
        begin = self.image_begin.reshape(1, 1, -1).expand(b, 1, x.shape[-1]).to(dtype)
        end = self.image_end.reshape(1, 1, -1).expand(b, 1, x.shape[-1]).to(dtype)
        x = torch.cat([begin, x, end], dim=1)
        x = self.after_rms(x)
        return x

    @classmethod
    def from_config(cls, model_config, running_config=None):
        return cls(
            model_config,
        )


str2adapter = {
    "llava": VisionLanguageAdapter,
    "gemma3": Gemma3MultiModalProjector,
    "deepseekocr": DeepSeekOCRProjector,
    "hunyuanocr": HunYuanVisionPatchMerger,
}



"""Implementation of the CNN Decoder part of
"Convolutional Sequence to Sequence Learning"
"""

import torch
import torch.nn as nn

from eole.modules.conv_multi_step_attention import ConvMultiStepAttention
from eole.utils.cnn_factory import shape_transform, GatedConv
from eole.decoders.decoder import DecoderBase

SCALE_WEIGHT = 0.5**0.5


class CNNDecoder(DecoderBase):
    """Decoder based on "Convolutional Sequence to Sequence Learning"
    :cite:`DBLP:journals/corr/GehringAGYD17`.

    Consists of residual convolutional layers, with ConvMultiStepAttention.
    """

    def __init__(
        self,
        decoder_config,
        running_config=None,
        with_cross_attn=False,
    ):
        super(CNNDecoder, self).__init__()

        self.cnn_kernel_width = decoder_config.cnn_kernel_width

        # Decoder State
        self.state = {}

        input_size = decoder_config.hidden_size  # we need embeddings.src_vec_size
        self.linear = nn.Linear(input_size, decoder_config.hidden_size)
        self.conv_layers = nn.ModuleList(
            [
                GatedConv(
                    decoder_config.hidden_size,
                    decoder_config.cnn_kernel_width,
                    getattr(running_config, "dropout", [0.0])[0],
                    True,
                )
                for i in range(decoder_config.layers)
            ]
        )
        self.attn_layers = nn.ModuleList(
            [ConvMultiStepAttention(decoder_config.hidden_size) for i in range(decoder_config.layers)]
        )

    @classmethod
    def from_config(cls, decoder_config, running_config=None, with_cross_attn=False):
        """Alternate constructor."""
        return cls(
            decoder_config,
            running_config,
            with_cross_attn=False,
        )

    def init_state(self, **kwargs):
        """Init decoder state."""
        enc_out = kwargs.pop("enc_out", None)
        enc_final_hs = kwargs.pop("enc_final_hs", None)
        self.state["src"] = (enc_out + enc_final_hs) * SCALE_WEIGHT
        self.state["previous_input"] = None

    def map_state(self, fn):
        self.state["src"] = fn(self.state["src"])
        if self.state["previous_input"] is not None:
            self.state["previous_input"] = fn(self.state["previous_input"])

    def detach_state(self):
        self.state["previous_input"] = self.state["previous_input"].detach()

    def forward(self, emb, enc_out, step=None, **kwargs):
        """See :obj:`eole.modules.RNNDecoderBase.forward()`"""

        if self.state["previous_input"] is not None:
            emb = torch.cat([self.state["previous_input"], emb], 1)

        dec_outs = []
        attns = {"std": []}

        assert emb.dim() == 3  # batch x len x embedding_dim

        tgt_emb = emb
        # The output of CNNEncoder.
        enc_out_t = enc_out
        # The combination of output of CNNEncoder and source embeddings.
        enc_out_c = self.state["src"]

        emb_reshape = tgt_emb.view(tgt_emb.size(0) * tgt_emb.size(1), -1)
        linear_out = self.linear(emb_reshape)
        x = linear_out.view(tgt_emb.size(0), tgt_emb.size(1), -1)
        x = shape_transform(x)

        pad = torch.zeros(x.size(0), x.size(1), self.cnn_kernel_width - 1, 1)

        pad = pad.type_as(x)
        base_target_emb = x

        for conv, attention in zip(self.conv_layers, self.attn_layers):
            new_target_input = torch.cat([pad, x], 2)
            out = conv(new_target_input)
            c, attn = attention(base_target_emb, out, enc_out_t, enc_out_c)
            x = (x + (c + out) * SCALE_WEIGHT) * SCALE_WEIGHT

        dec_outs = x.squeeze(3).transpose(1, 2)

        # Process the result and update the attentions.
        if self.state["previous_input"] is not None:
            dec_outs = dec_outs[:, self.state["previous_input"].size(1) :, :]
            attn = attn[:, self.state["previous_input"].size(1) :].squeeze()
            attn = torch.stack([attn])
        attns["std"] = attn

        # Update the state.
        self.state["previous_input"] = emb
        # TODO change the way attns is returned dict => list or tuple (onnx)
        return dec_outs, attns

    def update_dropout(self, dropout, attention_dropout=None):
        for layer in self.conv_layers:
            layer.dropout.p = dropout



#!/usr/bin/env python
import json
import torch
import pyonmttok
from eole.inputters.inputter import vocabs_to_dict
import os
from safetensors.torch import save_file

from eole.config.models import (
    EmbeddingsConfig,
    TransformerLMModelConfig,
)
from eole.config.run import TrainConfig
from eole.config.training import TrainingConfig
from eole.config import recursive_model_fields_set
from eole.bin import BaseBin, register_bin


@register_bin(name="mpt")
class MPTConverter(BaseBin):
    @classmethod
    def add_args(cls, parser):
        parser.add_argument(
            "--model_dir",
            type=str,
            required=True,
            help="""Path to the model directory""",
        )
        parser.add_argument(
            "--vocab_file",
            type=str,
            required=True,
            help="""Path to the tokenizer model""",
        )
        parser.add_argument("--output", type=str, required=True, help="""Path to the model directory""")
        parser.add_argument("--nshards", type=int, default=1, help="""Path to the model directory""")

    @classmethod
    def run(cls, args):
        print("Loading the full model in Cpu Ram ...")
        from transformers import AutoModelForCausalLM

        model = AutoModelForCausalLM.from_pretrained(
            args.model_dir,
            torch_dtype=torch.float16,
            device_map={"": "cpu"},
            trust_remote_code=True,
        )
        checkpoint = model.state_dict()

        params_json = os.path.join(args.model_dir, "config.json")
        with open(params_json, encoding="utf-8") as fparam:
            params = json.load(fparam)

        decoder_layers = params["n_layers"]
        src_word_vec_size = params["d_model"]
        tgt_word_vec_size = params["d_model"]
        hidden_size = params["d_model"]
        heads = params["n_heads"]
        vocab_size = params["vocab_size"]
        transformer_ff = params["expansion_ratio"] * params["d_model"]

        for shard in range(args.nshards):

            print("starting output shard: %d/%d" % (shard + 1, args.nshards))
            eole_safetensor = {}

            if shard == 0:
                eole_safetensor["tgt_emb.embeddings.weight"] = checkpoint["transformer.wte.weight"]
                eole_safetensor["decoder.layer_norm.weight"] = checkpoint["transformer.norm_f.weight"]
                eole_safetensor["decoder.layer_norm.bias"] = torch.zeros(
                    eole_safetensor["decoder.layer_norm.weight"].size(0),
                    dtype=torch.float16,
                )

                eole_safetensor["generator.weight"] = checkpoint["transformer.wte.weight"].clone()
                eole_safetensor["generator.bias"] = torch.zeros(
                    eole_safetensor["generator.weight"].size(0), dtype=torch.float16
                )

            for i in range(
                -(decoder_layers // -args.nshards) * shard,
                min(-(decoder_layers // -args.nshards) * (shard + 1), decoder_layers),
                1,
            ):
                eole_safetensor["decoder.transformer_layers." + str(i) + ".self_attn.linear_query.weight"] = checkpoint[
                    "transformer.blocks." + str(i) + ".attn.Wqkv.weight"
                ][:hidden_size, :].clone()
                eole_safetensor["decoder.transformer_layers." + str(i) + ".self_attn.linear_keys.weight"] = checkpoint[
                    "transformer.blocks." + str(i) + ".attn.Wqkv.weight"
                ][hidden_size : (hidden_size * 2), :].clone()
                eole_safetensor["decoder.transformer_layers." + str(i) + ".self_attn.linear_values.weight"] = (
                    checkpoint["transformer.blocks." + str(i) + ".attn.Wqkv.weight"][(hidden_size * 2) :, :].clone()
                )

                eole_safetensor["decoder.transformer_layers." + str(i) + ".self_attn.final_linear.weight"] = checkpoint[
                    "transformer.blocks." + str(i) + ".attn.out_proj.weight"
                ]

                eole_safetensor["decoder.transformer_layers." + str(i) + ".layer_norm_1.weight"] = checkpoint[
                    "transformer.blocks." + str(i) + ".norm_1.weight"
                ]
                eole_safetensor["decoder.transformer_layers." + str(i) + ".layer_norm_1.bias"] = torch.zeros(
                    eole_safetensor["decoder.transformer_layers." + str(i) + ".layer_norm_1.weight"].size(0),
                    dtype=torch.float16,
                )

                eole_safetensor["decoder.transformer_layers." + str(i) + ".feed_forward.w_1.weight"] = checkpoint[
                    "transformer.blocks." + str(i) + ".ffn.up_proj.weight"
                ]

                eole_safetensor["decoder.transformer_layers." + str(i) + ".feed_forward.w_2.weight"] = checkpoint[
                    "transformer.blocks." + str(i) + ".ffn.down_proj.weight"
                ]

                eole_safetensor["decoder.transformer_layers." + str(i) + ".feed_forward.layer_norm.weight"] = (
                    checkpoint["transformer.blocks." + str(i) + ".norm_2.weight"]
                )
                eole_safetensor["decoder.transformer_layers." + str(i) + ".feed_forward.layer_norm.bias"] = torch.zeros(
                    eole_safetensor["decoder.transformer_layers." + str(i) + ".feed_forward.layer_norm.weight"].size(0),
                    dtype=torch.float16,
                )

            if shard == 0:
                transformer_ff = eole_safetensor["decoder.transformer_layers.0.feed_forward.w_1.weight"].size(0)
                vocab_size = eole_safetensor["generator.weight"].size(0)
            print("Saving output model shard: %d" % shard)
            save_file(
                eole_safetensor,
                os.path.join(args.output, "model.{:02d}.safetensors".format(shard)),
            )

        vocabs = {}
        with open(args.vocab_file, "r", encoding="utf-8") as vocab:
            src_vocab = pyonmttok.build_vocab_from_lines(vocab)

        vocabs["src"] = src_vocab
        vocabs["tgt"] = src_vocab
        vocabs["decoder_start_token"] = "</s>"

        vocab_dict = vocabs_to_dict(vocabs)
        with open(os.path.join(args.output, "vocab.json"), "w", encoding="utf-8") as f:
            json.dump(vocab_dict, f, indent=2, ensure_ascii=False)

        with open(os.path.join(args.output, "vocab.txt"), "w", encoding="utf-8") as vocabfile:
            for tok in vocab_dict["src"]:
                vocabfile.write(tok + "\n")

        position_encoding = {
            "position_encoding_type": "Alibi",
            "n_positions": 0,
        }

        config = TrainConfig(
            data=None,
            skip_empty_level="silent",  # default is "warning"
            save_data=None,
            n_sample=0,
            src_vocab=None,
            tgt_vocab=None,
            share_vocab=True,
            src_vocab_size=vocab_size,
            tgt_vocab_size=vocab_size,
            vocab_size_multiple=8,
            decoder_start_token=vocabs["decoder_start_token"],
            transforms=["filtertoolong"],
            transforms_configs={"filtertoolong": {"src_seq_length": 512, "tgt_seq_length": 512}},
            model=TransformerLMModelConfig(
                layers=decoder_layers,
                hidden_size=hidden_size,
                heads=heads,
                transformer_ff=transformer_ff,
                embeddings=EmbeddingsConfig(
                    src_word_vec_size=src_word_vec_size,
                    tgt_word_vec_size=tgt_word_vec_size,
                    **position_encoding,
                ),
                # src_word_vec_size=src_word_vec_size,
                # tgt_word_vec_size=tgt_word_vec_size,
                layer_norm="standard",
                pos_ffn_activation_fn="gelu",
                parallel_residual=False,
                add_qkvbias=False,
                add_ffnbias=False,
            ),
            training=TrainingConfig(
                compute_dtype="fp16",
                batch_size=896,
                batch_size_multiple=1,
                batch_type="tokens",
                normalization="tokens",
                accum_count=[32],
                accum_steps=[0],
                valid_batch_size=256,
            ),
        )

        config_dict = recursive_model_fields_set(config)
        with open(os.path.join(args.output, "config.json"), "w", encoding="utf-8") as f:
            json.dump(config_dict, f, indent=2, ensure_ascii=False)



from eole.utils.logging import logger
from eole.transforms import register_transform
from .transform import Transform, TransformConfig
from pydantic import Field
import random
import string


class InlineTagsConfig(TransformConfig):
    tags_dictionary_path: str | None = Field(default=None, description="Path to a flat term dictionary.")
    tags_corpus_ratio: float | None = Field(default=0.1, description="Ratio of corpus to augment with tags.")
    max_tags: int | None = Field(
        default=12,
        description="Maximum number of tags that " "can be added to a single sentence.",
    )
    paired_stag: str | None = Field(
        default="｟ph_#_beg｠",
        description="The format of an opening paired inline tag. " "Must include the character #.",
    )
    paired_etag: str | None = Field(
        default="｟ph_#_end｠",
        description="The format of a closing paired inline tag. " "Must include the character #.",
    )
    isolated_tag: str | None = Field(
        default="｟ph_#_std｠",
        description="The format of an isolated inline tag. " "Must include the character #.",
    )
    src_delimiter: str | None = Field(
        default="｟fuzzy｠",
        description="Any special token used for augmented src sentences. "
        "The default is the fuzzy token used in the FuzzyMatch transform.",
    )


class InlineTagger(object):
    """Class for augmenting source and target sentences
    with inline tags (placeholders).

    It requires a prepared tab-delimited dictionary of source-target
    words and phrases. A dictionary can be created with
    tools such as fastalign and it should ideally contain enough long
    phrases (of 3-4 words or more) for more realistic applications
    of start and end tags. A dictionary with 20.000-30.000 entries
    should give sufficient number of matches."""

    def __init__(
        self,
        tags_dictionary_path,
        max_tags,
        paired_start_tag,
        paired_end_tag,
        isolated_tag,
        src_delimiter,
        tag_corpus_ratio=0.1,
    ):
        self.max_tags = max_tags
        self.tag_corpus_ratio = tag_corpus_ratio
        self.src_delimiter = src_delimiter
        self.internal_dictionary = self._create_internal_dictionary(tags_dictionary_path)
        self.paired_stag_prefix, self.paired_stag_suffix = map(str, paired_start_tag.split("#"))
        self.paired_etag_prefix, self.paired_etag_suffix = map(str, paired_end_tag.split("#"))
        self.isolated_tag_prefix, self.isolated_tag_suffix = map(str, isolated_tag.split("#"))

        self.automaton = self._create_automaton()

    def _create_internal_dictionary(self, tags_dictionary_path):
        dictionary = list()
        with open(tags_dictionary_path, mode="r", encoding="utf-8") as file:
            pairs = file.readlines()
            for pair in pairs:
                src_term, tgt_term = map(str, pair.split("\t"))
                dictionary.append((src_term.strip(), tgt_term.strip()))
        return dictionary

    def _create_automaton(self):
        import ahocorasick

        automaton = ahocorasick.Automaton()
        for entry in self.internal_dictionary:
            automaton.add_word(entry[0], (entry[0], entry[1]))

        automaton.make_automaton()
        return automaton

    def _tagged_src_tgt(self, src_example, tgt_example) -> tuple:
        """Uses the dictionary to find exact source matches with corresponding
        target matches and adds both paired tags and standalone tags."""

        maybe_augmented = src_example.split(self.src_delimiter)
        source_only = maybe_augmented[0].strip()

        augmented_part = maybe_augmented[1].strip() if len(maybe_augmented) > 1 else None

        tokenized_source_string = source_only.split(" ")
        tokenized_target_string = tgt_example.split(" ")

        src_offset, tgt_offset = 0, 0
        src_with_tags, tgt_with_tags = list(), list()

        # We set the start number of tags to a random number from 1
        # to 12 + the number of subsequent tags that
        # will be added. We also apply weights to this choice so tags
        # are more probable to start from 1, then from 2, etc.
        # This way we cover most scenarios met in real usage and
        # the system will learn to handle a fairly large number of
        # numbered tags (but not an excessively large number)
        paired_tag_start_num = random.choices(range(1, self.max_tags + 1), weights=range(self.max_tags, 0, -1), k=1)[0]
        single_tag_start_num = random.choices(range(1, self.max_tags + 1), weights=range(self.max_tags, 0, -1), k=1)[0]

        is_match = False
        tag_counter = 0

        for src_match_end, pair in self.automaton.iter_long(source_only):
            if tag_counter == self.max_tags:
                break

            src_match_start = src_match_end - len(pair[0]) + 1
            tgt_match_start = tgt_example.find(pair[1], tgt_offset)
            tgt_match_end = tgt_match_start + len(pair[1])

            # Make sure we only search for exact matches (we don't want
            # to match part of words) and perform some bound checking
            if (
                (pair[1] not in " ".join(tokenized_target_string))
                or (
                    len(source_only) != src_match_end + 1
                    and not (
                        source_only[src_match_end + 1].isspace() or source_only[src_match_end + 1] in string.punctuation
                    )
                )
                or (not source_only[src_match_start - 1].isspace() and src_match_start != 0)
            ):
                continue
            else:
                source_index = 0
                for i, w in enumerate(tokenized_source_string):
                    if source_index == src_match_start:
                        source_index = i
                        break
                    else:
                        source_index += len(w) + 1

                target_index = 0
                for i, w in enumerate(tokenized_target_string):
                    if target_index == tgt_match_start:
                        target_index = i
                        break
                    else:
                        target_index += len(w) + 1

                src_term = " ".join(tokenized_source_string[source_index : source_index + len(pair[0].split(" "))])
                tgt_term = " ".join(tokenized_target_string[target_index : target_index + len(pair[1].split(" "))])

                # Create all possible tag forms. We inject a special
                # unicode char (∥) as a placeholder for whitespace in order
                # to keep the indices unaltered. This char is replaced with
                # spaces before we return the augmented examples.
                src_single_tags = (
                    f"{source_only[src_offset: src_match_start]}"
                    f"{self.isolated_tag_prefix}{single_tag_start_num}"
                    f"{self.isolated_tag_suffix}∥{src_term}∥"
                )
                src_paired_tags = (
                    f"{source_only[src_offset: src_match_start]}"
                    f"{self.paired_stag_prefix}{paired_tag_start_num}"
                    f"{self.paired_stag_suffix}∥{src_term}∥"
                    f"{self.paired_etag_prefix}{paired_tag_start_num}"
                    f"{self.paired_etag_suffix}"
                )

                tgt_single_tags = (
                    f"{tgt_example[tgt_offset: tgt_match_start]}"
                    f"{self.isolated_tag_prefix}{single_tag_start_num}"
                    f"{self.isolated_tag_suffix}∥{tgt_term}∥"
                )
                tgt_paired_tags = (
                    f"{tgt_example[tgt_offset: tgt_match_start]}"
                    f"{self.paired_stag_prefix}{paired_tag_start_num}"
                    f"{self.paired_stag_suffix}∥{tgt_term}∥"
                    f"{self.paired_etag_prefix}{paired_tag_start_num}"
                    f"{self.paired_etag_suffix}∥"
                )

                # Make a weighted choice between paired tags or single tags.
                # We usually encounter, and thus here we favor, paired tags
                # with a ratio 1/3.
                choice = random.choices([src_single_tags, src_paired_tags], weights=(1, 3), k=1)

                src_with_tags.append(choice[0])
                src_offset = src_match_end + 1

                if choice[0] is src_single_tags:
                    tgt_with_tags.append(tgt_single_tags)
                    single_tag_start_num += 1
                else:
                    tgt_with_tags.append(tgt_paired_tags)
                    paired_tag_start_num += 1

                tgt_offset = tgt_match_end + 1
                tag_counter += 1
                is_match = True
        if is_match:
            if augmented_part is not None:
                src_with_tags.append(source_only[src_offset:] + self.src_delimiter + augmented_part)
            else:
                src_with_tags.append(source_only[src_offset:])

            tgt_with_tags.append(tgt_example[tgt_offset:])

            return (
                "".join(src_with_tags).replace("∥", " ").split(" "),
                "".join(tgt_with_tags).replace("∥", " ").split(" "),
            ), is_match
        else:
            return (src_example.split(" "), tgt_example.split(" ")), is_match


@register_transform(name="inlinetags")
class InlineTagsTransform(Transform):

    config_model = InlineTagsConfig

    def __init__(self, config):
        super().__init__(config)

    def _parse_config(self):
        self.tags_dictionary_path = self.config.tags_dictionary_path
        self.tags_corpus_ratio = self.config.tags_corpus_ratio
        self.max_tags = self.config.max_tags
        self.src_delimiter = self.config.src_delimiter

    @classmethod
    def get_specials(cls, config):
        """Add up to self.max_tags * 2 placeholders to src and tgt vocabs."""

        # Check if the tags include the
        # mandatory "#" number placeholder"
        if "#" not in config.paired_stag or "#" not in config.paired_etag or "#" not in config.isolated_tag:
            logger.error("Inline tags must include the number " 'placeholder "#"')

        # We split the user-defined tags in the # placeholder
        # in order to number them
        paired_stag_prefix, paired_stag_suffix = map(str, config.paired_stag.split("#"))
        paired_etag_prefix, paired_etag_suffix = map(str, config.paired_etag.split("#"))
        isolated_tag_prefix, isolated_tag_suffix = map(str, config.isolated_tag.split("#"))

        src_specials, tgt_specials = list(), list()
        tags = list()
        for i in range(1, config.max_tags * 2):
            tags.extend(
                [
                    paired_stag_prefix + str(i) + paired_stag_suffix,
                    paired_etag_prefix + str(i) + paired_etag_suffix,
                    isolated_tag_prefix + str(i) + isolated_tag_suffix,
                ]
            )

        src_specials.extend(tags)
        tgt_specials.extend(tags)

        return (src_specials, tgt_specials)

    def warm_up(self, vocabs=None):
        """Create the tagger."""

        super().warm_up(None)
        self.tagger = InlineTagger(
            self.tags_dictionary_path,
            self.max_tags,
            self.config.paired_stag,
            self.config.paired_etag,
            self.config.isolated_tag,
            self.src_delimiter,
            self.tags_corpus_ratio,
        )

    def batch_apply(self, batch, is_train=False, stats=None, **kwargs):
        bucket_size = len(batch)
        examples_with_tags = 0

        for i, (ex, _, _) in enumerate(batch):
            # Skip half examples to speed up the transform. This sets
            # a hard limit of 0.5 to the `tags_corpus_ratio`, which is
            # excessive and should be avoided anyway.
            if i % 2 == 0:
                original_src = ex["src"]
                original_tgt = ex["tgt"]
                augmented_example, is_match = self.apply(ex, is_train, stats, **kwargs)
                if is_match and (examples_with_tags < bucket_size * self.tags_corpus_ratio):
                    examples_with_tags += 1
                    ex["src"] = augmented_example["src"]
                    ex["tgt"] = augmented_example["tgt"]
                else:
                    ex["src"] = original_src
                    ex["tgt"] = original_tgt

        logger.debug(f"Added tags to {examples_with_tags}/{bucket_size} examples")
        return batch

    def apply(self, example, is_train=False, stats=None, **kwargs) -> tuple:
        """Add tags (placeholders) to source and target segments."""

        src_tgt_pair, is_match = self.tagger._tagged_src_tgt(" ".join(example["src"]), " ".join(example["tgt"]))
        example["src"] = src_tgt_pair[0]
        example["tgt"] = src_tgt_pair[1]

        return example, is_match



#!/usr/bin/env python
""" Translator Class and builder """
import torch
from torch.nn.utils.rnn import pad_sequence
from eole.predict.beam_search import BeamSearch
from eole.predict.greedy_search import GreedySearch
from eole.utils.misc import tile
from eole.utils.alignment import extract_alignment
from eole.predict.inference import Inference
from time import time


class Translator(Inference):
    def _align_forward(self, batch, predictions):
        """
        For a batch of input and its prediction, return a list of batch predict
        alignment src indice Tensor in size ``(batch, n_best,)``.
        """

        # (0) add BOS and padding to tgt prediction
        if "tgt" in batch.keys() and self.gold_align:
            self._log("Computing alignments with gold target")
            batch_tgt_idxs = batch["tgt"].transpose(1, 2)
        else:
            batch_tgt_idxs = self._align_pad_prediction(predictions, bos=self._tgt_bos_idx, pad=self._tgt_pad_idx)

        tgt_mask = (
            batch_tgt_idxs.eq(self._tgt_pad_idx)
            | batch_tgt_idxs.eq(self._tgt_bos_idx)
            | torch.any(
                torch.stack([batch_tgt_idxs.eq(eos_idx) for eos_idx in self._tgt_eos_idx]),
                dim=0,
            )
        )

        n_best = batch_tgt_idxs.size(1)
        # (1) Encoder forward.
        src, enc_states, enc_out, src_len = self._run_encoder(batch)

        # (2) Repeat src objects `n_best` times.
        # We use batch_size x n_best, get ``(batch * n_best, src_len)``
        src = tile(src, n_best)
        if enc_states is not None:
            # Quick fix. Transformers return None as enc_states.
            # enc_states are only used later on to init decoder's state
            # but are never used in Transformer decoder, so we can skip
            enc_states = tile(enc_states, n_best)
        if isinstance(enc_out, tuple):
            enc_out = tuple(tile(x, n_best) for x in enc_out)
        else:
            enc_out = tile(enc_out, n_best)
        src_len = tile(src_len, n_best)  # ``(batch * n_best,)``

        # (3) Init decoder with n_best src,
        self.model.decoder.init_state(src=src, enc_out=enc_out, enc_states=enc_states)

        # (4) reshape and apply pad masking in the target sequence
        tgt = batch_tgt_idxs.view(-1, batch_tgt_idxs.size(-1))
        src_pad_mask = src.eq(self._src_pad_idx).unsqueeze(1)
        tgt_pad_mask = tgt[:, :-1].eq(self._tgt_pad_idx).unsqueeze(1)
        dec_in = tgt[:, :-1]
        _, attns = self.model.decoder(
            self.model.tgt_emb(dec_in),
            enc_out=enc_out,
            src_pad_mask=src_pad_mask,
            tgt_pad_mask=tgt_pad_mask,
            with_align=True,
        )

        alignment_attn = attns["align"]  # ``(B, tgt_len-1, src_len)``
        # masked_select
        align_tgt_mask = tgt_mask.view(-1, tgt_mask.size(-1))
        prediction_mask = align_tgt_mask[:, 1:]  # exclude bos to match pred
        # get aligned src id for each prediction's valid tgt tokens
        alignment = extract_alignment(alignment_attn, prediction_mask, src_len, n_best)
        return alignment

    def predict_batch(self, batch, attn_debug):
        """Translate a batch of sentences."""
        if self.max_length_ratio > 0:
            max_length = int(min(self.max_length, batch["src"].size(1) * self.max_length_ratio + 5))
        else:
            max_length = self.max_length
        with torch.no_grad():
            if self.top_k != 0 or self.top_p != 0:
                decode_strategy = GreedySearch(
                    pad=self._tgt_pad_idx,
                    bos=self._tgt_bos_idx,
                    eos=self._tgt_eos_idx,
                    unk=self._tgt_unk_idx,
                    start=self._tgt_start_with,
                    n_best=self.n_best,
                    batch_size=len(batch["srclen"]),
                    global_scorer=self.global_scorer,
                    min_length=self.min_length,
                    max_length=max_length,
                    block_ngram_repeat=self.block_ngram_repeat,
                    exclusion_tokens=self._exclusion_idxs,
                    return_attention=attn_debug or self.replace_unk,
                    temperature=self.temperature,
                    top_k=self.top_k,
                    top_p=self.top_p,
                    beam_size=self.beam_size,
                    ban_unk_token=self.ban_unk_token,
                    add_estimator=self.add_estimator,
                )
            else:
                # TODO: support these blacklisted features
                assert not self.dump_beam
                decode_strategy = BeamSearch(
                    self.beam_size,
                    batch_size=len(batch["srclen"]),
                    pad=self._tgt_pad_idx,
                    bos=self._tgt_bos_idx,
                    eos=self._tgt_eos_idx,
                    unk=self._tgt_unk_idx,
                    start=self._tgt_start_with,
                    n_best=self.n_best,
                    global_scorer=self.global_scorer,
                    min_length=self.min_length,
                    max_length=max_length,
                    return_attention=attn_debug or self.replace_unk,
                    block_ngram_repeat=self.block_ngram_repeat,
                    exclusion_tokens=self._exclusion_idxs,
                    stepwise_penalty=self.stepwise_penalty,
                    ratio=self.ratio,
                    ban_unk_token=self.ban_unk_token,
                    add_estimator=self.add_estimator,
                )
            return self._predict_batch_with_strategy(batch, decode_strategy)

    def _run_encoder(self, batch):
        src = batch["src"]
        src_len = batch["srclen"]
        batch_size = len(batch["srclen"])
        emb = self.model.src_emb(src)
        pad_mask = src.eq(self._src_pad_idx).unsqueeze(1)  # [B, 1, T_src]
        enc_out, enc_final_hs = self.model.encoder(emb, pad_mask=pad_mask)

        if src_len is None:
            assert not isinstance(enc_out, tuple), "Ensemble decoding only supported for text data"
            src_len = torch.Tensor(batch_size).type_as(enc_out).long().fill_(enc_out.size(1))
        return src, enc_final_hs, enc_out, src_len

    def _predict_batch_with_strategy(self, batch, decode_strategy):
        """Translate a batch of sentences step by step using cache.

        Args:
            batch: a batch of sentences, yield by data iterator.
            decode_strategy (DecodeStrategy): A decode strategy to use for
                generate translation step by step.

        Returns:
            results (dict): The translation results.
        """

        # (0) Prep the components of the search.
        parallel_paths = decode_strategy.parallel_paths  # beam_size

        batch_size = len(batch["srclen"])

        # (1) Run the encoder on the src.
        if self.report_time:
            torch.cuda.synchronize()
            beg_time = time()
        src, enc_final_hs, enc_out, src_len = self._run_encoder(batch)
        self.model.decoder.init_state(src=src, enc_out=enc_out, enc_final_hs=enc_final_hs)

        gold_score, gold_log_probs = self._gold_score(
            batch,
            enc_out,
            src_len,
            enc_final_hs,
            batch_size,
            src,
        )

        # (2) prep decode_strategy. Possibly repeat src objects.
        target_prefix = batch["tgt"] if self.tgt_file_prefix else None
        (fn_tile, enc_out) = decode_strategy.initialize(enc_out, src_len, target_prefix=target_prefix)

        if fn_tile is not None:
            self.model.decoder.map_state(fn_tile)

        # (3) Begin decoding step by step:
        # save the initial encoder out for later estimator
        if torch.is_tensor(enc_out):
            enc_out2 = enc_out.clone()
            src_len2 = decode_strategy.src_len.clone()
        elif isinstance(enc_out, tuple):
            enc_out2 = tuple(tensor.clone() for tensor in enc_out)
            src_len2 = tuple(tensor.clone() for tensor in decode_strategy.src_len)
        else:
            raise TypeError("enc_out must be either a tensor or a tuple of tensors")

        for step in range(decode_strategy.max_length):
            decoder_input = decode_strategy.current_predictions.view(-1, 1)

            log_probs, attn = self._decode_and_generate(
                decoder_input,
                enc_out,
                src_len=decode_strategy.src_len,
                step=step,
                return_attn=decode_strategy.return_attention,
            )

            decode_strategy.advance(log_probs, attn)
            any_finished = any([any(sublist) for sublist in decode_strategy.is_finished_list])
            if any_finished:
                decode_strategy.update_finished()
                if decode_strategy.done:
                    break

            select_indices = decode_strategy.select_indices

            if any_finished:
                # Reorder states.
                if isinstance(enc_out, tuple):
                    enc_out = tuple(x[select_indices] for x in enc_out)
                else:
                    enc_out = enc_out[select_indices]

            if parallel_paths > 1 or any_finished:
                self.model.decoder.map_state(lambda state: state[select_indices])

            if self.report_time and step == 0:
                torch.cuda.synchronize()
                self.step0_time.append(time() - beg_time)

        if self.add_estimator:
            dec_in = [item for sublist in decode_strategy.predictions for item in sublist]

            # Prepare estimator input = decoder out of each pred with initial enc_out
            dec_in = pad_sequence(dec_in, batch_first=True, padding_value=self._tgt_pad_idx)
            prepend_value = torch.full(
                (dec_in.size(0), 1),
                self._tgt_bos_idx,
                dtype=dec_in.dtype,
                device=dec_in.device,
            )
            dec_in = torch.cat((prepend_value, dec_in), dim=1)
            src_pad_mask = src.eq(self._src_pad_idx).unsqueeze(1)  # [B, 1, T_src]
            src_pad_mask = tile(src_pad_mask, parallel_paths)
            tgt_pad_mask = dec_in[:, :-1].eq(self._tgt_pad_idx).unsqueeze(1)  # [B, 1, T_tgt]
            emb = self.model.tgt_emb(dec_in[:, :-1])
            self.model.decoder._disable_cache()
            dec_out, _ = self.model.decoder(
                emb,
                enc_out=enc_out2,
                src_len=src_len2,
                step=None,
                return_attn=False,
                src_pad_mask=src_pad_mask,
                tgt_pad_mask=tgt_pad_mask,
            )
            if self.estimator_type == "average":
                pad_mask2 = ~dec_in[:, :-1].eq(self._tgt_pad_idx)
                in_estim2 = (dec_out * pad_mask2.unsqueeze(-1).float()).sum(dim=1) / pad_mask2.sum(
                    dim=1, keepdim=True
                ).float()
            elif self.estimator_type == "last_token":
                in_estim2 = dec_out[:, -1, :]
            else:
                raise ValueError("EncoderDecoder model should use average or last token estimator")
            estim = self.model.estimator(in_estim2.to(dec_out.dtype)).squeeze(-1)
        else:
            if torch.is_tensor(enc_out2):
                estim = torch.ones([enc_out2.size(0)])
            else:
                estim = torch.ones([enc_out2[0].size(0)])
        estim = [[estim[i].item() for i in range(j, j + self.beam_size)] for j in range(0, len(estim), self.beam_size)]

        return self.report_results(
            gold_score,
            gold_log_probs,
            batch,
            batch_size,
            decode_strategy,
            estim,
        )

    def _score_target(self, batch, enc_out, src_len):
        tgt = batch["tgt"]
        tgt_in = tgt[:, :-1]

        log_probs, attn = self._decode_and_generate(
            tgt_in,
            enc_out,
            src_len=src_len,
        )

        log_probs[:, :, self._tgt_pad_idx] = 0
        gold = tgt[:, 1:].unsqueeze(-1)
        gold_scores = log_probs.gather(2, gold)
        gold_scores = gold_scores.sum(dim=1).view(-1)
        return gold_scores, None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import division
import os
import six
import json
import torch
from eole.utils.logging import logger
from eole.inputters.inputter import dict_to_vocabs
from eole.bin import BaseBin, register_bin


def get_vocabs(model_path):
    # model = torch.load(dict_path, map_location=torch.device("cpu"))
    vocabs_path = os.path.join(model_path, "vocab.json")
    with open(vocabs_path, encoding="utf-8") as f:
        vocabs_dict = json.load(f)
    vocabs = dict_to_vocabs(vocabs_dict)

    enc_vocab = vocabs["src"]
    dec_vocab = vocabs["tgt"]

    logger.info("From: %s" % vocabs_path)
    logger.info("\t* source vocab: %d words" % len(enc_vocab))
    logger.info("\t* target vocab: %d words" % len(dec_vocab))

    return enc_vocab, dec_vocab


def read_embeddings(file_enc, skip_lines=0, filter_set=None):
    embs = dict()
    total_vectors_in_file = 0
    with open(file_enc, "rb") as f:
        for i, line in enumerate(f):
            if i < skip_lines:
                continue
            if not line:
                break
            if len(line) == 0:
                # is this reachable?
                continue

            l_split = line.decode("utf8").strip().split(" ")
            if len(l_split) == 2:
                continue
            total_vectors_in_file += 1
            if filter_set is not None and l_split[0] not in filter_set:
                continue
            embs[l_split[0]] = [float(em) for em in l_split[1:]]
    return embs, total_vectors_in_file


def convert_to_torch_tensor(word_to_float_list_dict, vocab):
    dim = len(six.next(six.itervalues(word_to_float_list_dict)))
    tensor = torch.zeros((len(vocab), dim))
    for word, values in word_to_float_list_dict.items():
        tensor[vocab[word]] = torch.Tensor(values)
    return tensor


def calc_vocab_load_stats(vocab, loaded_embed_dict):
    matching_count = len(set(vocab.tokens_to_ids.values()) & set(loaded_embed_dict.keys()))
    missing_count = len(vocab) - matching_count
    percent_matching = matching_count / len(vocab) * 100
    return matching_count, missing_count, percent_matching


@register_bin(name="embeddings_to_torch")
class EmbeddingsToTorch(BaseBin):
    @classmethod
    def add_args(cls, parser):
        parser.add_argument(
            "-emb_file_both",
            required=False,
            help="loads Embeddings for both source and target " "from this file.",
        )
        parser.add_argument("-emb_file_enc", required=False, help="source Embeddings from this file")
        parser.add_argument("-emb_file_dec", required=False, help="target Embeddings from this file")
        parser.add_argument("-output_file", required=True, help="Output file for the prepared data")
        parser.add_argument(
            "-model_path",
            required=True,
            help="Path to model directory with all components.",
        )
        parser.add_argument("-verbose", action="store_true", default=False)
        parser.add_argument(
            "-skip_lines",
            type=int,
            default=0,
            help="Skip first lines of the embedding file",
        )
        parser.add_argument("-type", choices=["GloVe", "word2vec"], default="GloVe")

    @classmethod
    def run(cls, args):
        enc_vocab, dec_vocab = get_vocabs(args.model_path)

        # Read in embeddings
        skip_lines = 1 if args.type == "word2vec" else args.skip_lines
        if args.emb_file_both is not None:
            if args.emb_file_enc is not None:
                raise ValueError("If --emb_file_both is passed in, you should not" "set --emb_file_enc.")
            if args.emb_file_dec is not None:
                raise ValueError("If --emb_file_both is passed in, you should not" "set --emb_file_dec.")
            set_of_src_and_tgt_vocab = set(enc_vocab.ids_to_tokens) | set(dec_vocab.ids_to_tokens)
            logger.info("Reading encoder and decoder embeddings from {}".format(args.emb_file_both))
            src_vectors, total_vec_count = read_embeddings(args.emb_file_both, skip_lines, set_of_src_and_tgt_vocab)
            tgt_vectors = src_vectors
            logger.info("\tFound {} total vectors in file".format(total_vec_count))
        else:
            if args.emb_file_enc is None:
                raise ValueError(
                    "If --emb_file_enc not provided. Please specify "
                    "the file with encoder embeddings, or pass in "
                    "--emb_file_both"
                )
            if args.emb_file_dec is None:
                raise ValueError(
                    "If --emb_file_dec not provided. Please specify "
                    "the file with encoder embeddings, or pass in "
                    "--emb_file_both"
                )
            logger.info("Reading encoder embeddings from {}".format(args.emb_file_enc))
            src_vectors, total_vec_count = read_embeddings(
                args.emb_file_enc, skip_lines, filter_set=set(enc_vocab.ids_to_tokens)
            )
            logger.info("\tFound {} total vectors in file.".format(total_vec_count))
            logger.info("Reading decoder embeddings from {}".format(args.emb_file_dec))
            tgt_vectors, total_vec_count = read_embeddings(
                args.emb_file_dec, skip_lines, filter_set=set(dec_vocab.ids_to_tokens)
            )
            logger.info("\tFound {} total vectors in file".format(total_vec_count))
        logger.info("After filtering to vectors in vocab:")
        logger.info("\t* enc: %d match, %d missing, (%.2f%%)" % calc_vocab_load_stats(enc_vocab, src_vectors))
        logger.info("\t* dec: %d match, %d missing, (%.2f%%)" % calc_vocab_load_stats(dec_vocab, tgt_vectors))

        # Write to file
        enc_output_file = args.output_file + ".enc.pt"
        dec_output_file = args.output_file + ".dec.pt"
        logger.info("\nSaving embedding as:\n\t* enc: %s\n\t* dec: %s" % (enc_output_file, dec_output_file))
        torch.save(convert_to_torch_tensor(src_vectors, enc_vocab), enc_output_file)
        torch.save(convert_to_torch_tensor(tgt_vectors, dec_vocab), dec_output_file)
        logger.info("\nDone.")



import json
import time
import yaml
from eole.config.run import PredictConfig
from argparse import ArgumentParser
from eole.utils.misc import use_gpu, set_random_seed


def evaluate(config, engine, input_file, out, method, model_type=None):
    print("# input file", input_file)
    run_results = {}
    # Build the translator (along with the model)
    if engine == "eole":
        print("Inference with py ...")
        from eole.inference_engine import InferenceEnginePY

        enginerunner = InferenceEnginePY(config)
    elif engine == "ct2":
        print("Inference with ct2 ...")
        from eole.inference_engine import InferenceEngineCT2

        config.src_subword_vocab = config.get_model_path() + "/vocabulary.json"
        enginerunner = InferenceEngineCT2(config, model_type=model_type)
    start = time.time()
    if method == "file":
        enginerunner.config.src = input_file
        scores, _, preds = enginerunner.infer_file()
    elif method == "list":
        src = open(input_file, ("r")).readlines()
        scores, _, preds = enginerunner.infer_list(src)
    enginerunner.terminate()
    dur = time.time() - start
    print(f"Time to generate {len(preds)} answers: {dur}s")
    if engine == "eole":
        scores = [[_score.cpu().numpy().tolist() for _score in _scores] for _scores in scores]
    run_results = {"pred_answers": preds, "score": scores, "duration": dur}
    output_filename = out + f"_{method}.json"
    with open(output_filename, "w") as f:
        json.dump(run_results, f, ensure_ascii=False, indent=2)


def main():
    # Required arguments
    parser = ArgumentParser()
    parser.add_argument("-model", help="Path to model.", required=True, type=str)
    parser.add_argument(
        "-model_type",
        help="Model task.",
        required=True,
        type=str,
        choices=["decoder", "encoder_decoder"],
    )
    parser.add_argument("-inference_config_file", help="Inference config file", required=True, type=str)
    parser.add_argument(
        "-engine",
        help="Engine",
        default="eole",
        required=False,
        type=str,
        choices=["eole", "ct2"],
    )
    parser.add_argument(
        "-input_file",
        help="File with formatted input examples.",
        required=True,
        type=str,
    )
    parser.add_argument(
        "-out",
        help="Output filename.",
        required=True,
        type=str,
    )
    args = parser.parse_args()
    model = args.model
    inference_config_file = args.inference_config_file
    with open(inference_config_file) as f:
        config = yaml.safe_load(f.read())
    config = PredictConfig(**config)
    set_random_seed(config.seed, use_gpu(config))
    config.model_path = [model]

    evaluate(
        config,
        engine=args.engine,
        input_file=args.input_file,
        out=args.out,
        method="file",
        model_type=args.model_type,
    )
    evaluate(
        config,
        engine=args.engine,
        input_file=args.input_file,
        out=args.out,
        method="list",
        model_type=args.model_type,
    )


if __name__ == "__main__":
    main()
